import numpy as np
from valor_lite.object_detection import DataLoader, Detection, Evaluator
from valor_lite.object_detection.computation import compute_confusion_matrix
from valor_lite.object_detection.utilities import _convert_example_to_dict


def test_confusion_matrix_no_data():
    evaluator = Evaluator()
    curves = evaluator.compute_confusion_matrix(
        iou_thresholds=[0.5],
        score_thresholds=[0.5],
        number_of_examples=0,
    )
    assert isinstance(curves, list)
    assert len(curves) == 0


def _test_compute_confusion_matrix(
    n_examples: int,
):
    sorted_pairs = np.array(
        [
            # dt,  gt,  pd,  iou,  gl,  pl, score,
            [0.0, 0.0, 1.0, 0.98, 0.0, 0.0, 0.9],
            [1.0, 1.0, 2.0, 0.55, 1.0, 0.0, 0.9],
            [2.0, -1.0, 4.0, 0.0, -1.0, 0.0, 0.65],
            [3.0, 4.0, 5.0, 1.0, 0.0, 0.0, 0.1],
            [1.0, 2.0, 3.0, 0.55, 0.0, 0.0, 0.1],
            [4.0, 5.0, -1.0, 0.0, 0.0, -1.0, -1.0],
        ]
    )
    label_metadata = np.array([[3, 4], [1, 0]])
    iou_thresholds = np.array([0.5])
    score_thresholds = np.array([score / 100.0 for score in range(1, 101)])

    (
        confusion_matrix,
        hallucinations,
        missing_predictions,
    ) = compute_confusion_matrix(
        data=sorted_pairs,
        label_metadata=label_metadata,
        iou_thresholds=iou_thresholds,
        score_thresholds=score_thresholds,
        n_examples=n_examples,
    )

    assert confusion_matrix.shape == (
        1,
        100,
        2,
        2,
        1 + n_examples * 4,
    )  # iou, score, gt label, pd label, metrics
    assert hallucinations.shape == (
        1,
        100,
        2,
        1 + n_examples * 3,
    )  # iou, score, pd label, metrics
    assert missing_predictions.shape == (
        1,
        100,
        2,
        1 + n_examples * 2,
    )  # iou, score, gt label, metrics

    return (confusion_matrix, hallucinations, missing_predictions)


def test_compute_confusion_matrix():

    (
        confusion_matrix,
        hallucinations,
        missing_predictions,
    ) = _test_compute_confusion_matrix(n_examples=0)

    """
    @ iou=0.5, score<0.1
    3x tp
    1x fp misclassification
    1x fp hallucination
    0x fn misclassification
    1x fn missing prediction
    """

    indices = slice(10)

    cm_gt0_pd0 = np.array([3.0])
    cm_gt0_pd1 = np.array([-1.0])
    cm_gt1_pd0 = np.array([1.0])
    cm_gt1_pd1 = np.array([-1.0])
    expected_cm = np.array(
        [[cm_gt0_pd0, cm_gt0_pd1], [cm_gt1_pd0, cm_gt1_pd1]]
    )
    assert np.isclose(confusion_matrix[0, indices, :, :, :], expected_cm).all()

    hal_pd0 = np.array([1.0])
    hal_pd1 = np.array([-1.0])
    expected_hallucinations = np.array([hal_pd0, hal_pd1])
    assert np.isclose(
        hallucinations[0, indices, :, :], expected_hallucinations
    ).all()

    misprd_gt0 = np.array([1.0])
    misprd_gt1 = np.array([-1.0])
    expected_missing_predictions = np.array([misprd_gt0, misprd_gt1])
    assert np.isclose(
        missing_predictions[0, indices, :, :], expected_missing_predictions
    ).all()

    """
    @ iou=0.5, 0.1 <= score < 0.65
    1x tp
    1x fp misclassification
    1x fp hallucination
    1x fn misclassification
    3x fn missing prediction
    """

    indices = slice(10, 65)

    cm_gt0_pd0 = np.array([1.0])
    cm_gt0_pd1 = np.array([-1.0])
    cm_gt1_pd0 = np.array([1.0])
    cm_gt1_pd1 = np.array([-1.0])
    expected_cm = np.array(
        [[cm_gt0_pd0, cm_gt0_pd1], [cm_gt1_pd0, cm_gt1_pd1]]
    )
    assert np.isclose(confusion_matrix[0, indices, :, :, :], expected_cm).all()

    hal_pd0 = np.array([1.0])
    hal_pd1 = np.array([-1.0])
    expected_hallucinations = np.array([hal_pd0, hal_pd1])
    assert np.isclose(
        hallucinations[0, indices, :, :], expected_hallucinations
    ).all()

    misprd_gt0 = np.array([3.0])
    misprd_gt1 = np.array([-1.0])
    expected_missing_predictions = np.array([misprd_gt0, misprd_gt1])
    assert np.isclose(
        missing_predictions[0, indices, :, :], expected_missing_predictions
    ).all()

    """
    @ iou=0.5, 0.65 <= score < 0.9
    1x tp
    1x fp misclassification
    0x fp hallucination
    1x fn misclassification
    3x fn missing prediction
    """

    indices = slice(65, 90)

    cm_gt0_pd0 = np.array([1.0])
    cm_gt0_pd1 = np.array([-1.0])
    cm_gt1_pd0 = np.array([1.0])
    cm_gt1_pd1 = np.array([-1.0])
    expected_cm = np.array(
        [[cm_gt0_pd0, cm_gt0_pd1], [cm_gt1_pd0, cm_gt1_pd1]]
    )
    assert np.isclose(confusion_matrix[0, indices, :, :, :], expected_cm).all()

    hal_pd0 = np.array([-1.0])
    hal_pd1 = np.array([-1.0])
    expected_hallucinations = np.array([hal_pd0, hal_pd1])
    assert np.isclose(
        hallucinations[0, indices, :, :], expected_hallucinations
    ).all()

    misprd_gt0 = np.array([3.0])
    misprd_gt1 = np.array([-1.0])
    expected_missing_predictions = np.array([misprd_gt0, misprd_gt1])
    assert np.isclose(
        missing_predictions[0, indices, :, :], expected_missing_predictions
    ).all()

    """
    @ iou=0.5, score>=0.9
    0x tp
    0x fp misclassification
    0x fp hallucination
    0x fn misclassification
    4x fn missing prediction
    """

    indices = slice(90, None)

    cm_gt0_pd0 = np.array([-1.0])
    cm_gt0_pd1 = np.array([-1.0])
    cm_gt1_pd0 = np.array([-1.0])
    cm_gt1_pd1 = np.array([-1.0])
    expected_cm = np.array(
        [[cm_gt0_pd0, cm_gt0_pd1], [cm_gt1_pd0, cm_gt1_pd1]]
    )
    assert np.isclose(confusion_matrix[0, indices, :, :, :], expected_cm).all()

    hal_pd0 = np.array([-1.0])
    hal_pd1 = np.array([-1.0])
    expected_hallucinations = np.array([hal_pd0, hal_pd1])
    assert np.isclose(
        hallucinations[0, indices, :, :], expected_hallucinations
    ).all()

    misprd_gt0 = np.array([4.0])
    misprd_gt1 = np.array([1.0])
    expected_missing_predictions = np.array([misprd_gt0, misprd_gt1])
    assert np.isclose(
        missing_predictions[0, indices, :, :], expected_missing_predictions
    ).all()


def test_compute_confusion_matrix_with_examples():

    (
        confusion_matrix,
        hallucinations,
        missing_predictions,
    ) = _test_compute_confusion_matrix(n_examples=2)

    """
    @ iou=0.5, score<0.1
    3x tp
    1x fp misclassification
    1x fp hallucination
    0x fn misclassification
    1x fn missing prediction
    """

    indices = slice(10)

    # total count, datum 0, gt 0, pd 0, score 0, datum 1, gt 1, pd 1, score 1
    cm_gt0_pd0 = np.array([3.0, 0.0, 0.0, 1.0, 0.9, 1.0, 2.0, 3.0, 0.1])
    cm_gt0_pd1 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    cm_gt1_pd0 = np.array([1.0, 1.0, 1.0, 2.0, 0.9, -1.0, -1.0, -1.0, -1.0])
    cm_gt1_pd1 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    expected_cm = np.array(
        [[cm_gt0_pd0, cm_gt0_pd1], [cm_gt1_pd0, cm_gt1_pd1]]
    )
    assert np.isclose(confusion_matrix[0, indices, :, :, :], expected_cm).all()

    # total count, datum 0, pd 0, score 0, datum 1, pd 1, score 1
    hal_pd0 = np.array([1.0, 2.0, 4.0, 0.65, -1.0, -1.0, -1.0])
    hal_pd1 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0])
    expected_hallucinations = np.array([hal_pd0, hal_pd1])
    assert np.isclose(
        hallucinations[0, indices, :, :], expected_hallucinations
    ).all()

    # total count, datum 0, gt 0, datum1, gt 1
    misprd_gt0 = np.array([1.0, 4.0, 5.0, -1.0, -1.0])
    misprd_gt1 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0])
    expected_missing_predictions = np.array([misprd_gt0, misprd_gt1])
    assert np.isclose(
        missing_predictions[0, indices, :, :], expected_missing_predictions
    ).all()

    """
    @ iou=0.5, 0.1 <= score < 0.65
    1x tp
    1x fp misclassification
    1x fp hallucination
    1x fn misclassification
    2x fn missing prediction
    """

    indices = slice(10, 65)

    # total count, datum 0, gt 0, pd 0, score 0, datum 1, gt 1, pd 1, score 1
    cm_gt0_pd0 = np.array([1.0, 0.0, 0.0, 1.0, 0.9, -1.0, -1.0, -1.0, -1.0])
    cm_gt0_pd1 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    cm_gt1_pd0 = np.array([1.0, 1.0, 1.0, 2.0, 0.9, -1.0, -1.0, -1.0, -1.0])
    cm_gt1_pd1 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    expected_cm = np.array(
        [[cm_gt0_pd0, cm_gt0_pd1], [cm_gt1_pd0, cm_gt1_pd1]]
    )
    assert np.isclose(confusion_matrix[0, indices, :, :, :], expected_cm).all()

    # total count, datum 0, pd 0, score 0, datum 1, pd 1, score 1
    hal_pd0 = np.array([1.0, 2.0, 4.0, 0.65, -1.0, -1.0, -1.0])
    hal_pd1 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0])
    expected_hallucinations = np.array([hal_pd0, hal_pd1])
    assert np.isclose(
        hallucinations[0, indices, :, :], expected_hallucinations
    ).all()

    # total count, datum 0, gt 0, datum1, gt 1
    misprd_gt0 = np.array([3.0, 1.0, 2.0, 3.0, 4.0])
    misprd_gt1 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0])
    expected_missing_predictions = np.array([misprd_gt0, misprd_gt1])
    assert np.isclose(
        missing_predictions[0, indices, :, :], expected_missing_predictions
    ).all()

    """
    @ iou=0.5, 0.65 <= score < 0.9
    1x tp
    1x fp misclassification
    0x fp hallucination
    1x fn misclassification
    2x fn missing prediction
    """

    indices = slice(65, 90)

    # total count, datum 0, gt 0, pd 0, score 0, datum 1, gt 1, pd 1, score 1
    cm_gt0_pd0 = np.array([1.0, 0.0, 0.0, 1.0, 0.9, -1.0, -1.0, -1.0, -1.0])
    cm_gt0_pd1 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    cm_gt1_pd0 = np.array([1.0, 1.0, 1.0, 2.0, 0.9, -1.0, -1.0, -1.0, -1.0])
    cm_gt1_pd1 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    expected_cm = np.array(
        [[cm_gt0_pd0, cm_gt0_pd1], [cm_gt1_pd0, cm_gt1_pd1]]
    )
    assert np.isclose(confusion_matrix[0, indices, :, :, :], expected_cm).all()

    # total count, datum 0, pd 0, score 0, datum 1, pd 1, score 1
    hal_pd0 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0])
    hal_pd1 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0])
    expected_hallucinations = np.array([hal_pd0, hal_pd1])
    assert np.isclose(
        hallucinations[0, indices, :, :], expected_hallucinations
    ).all()

    # total count, datum 0, gt 0, datum1, gt 1
    misprd_gt0 = np.array([3.0, 1.0, 2.0, 3.0, 4.0])
    misprd_gt1 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0])
    expected_missing_predictions = np.array([misprd_gt0, misprd_gt1])
    assert np.isclose(
        missing_predictions[0, indices, :, :], expected_missing_predictions
    ).all()

    """
    @ iou=0.5, score>=0.9
    0x tp
    0x fp misclassification
    0x fp hallucination
    0x fn misclassification
    4x fn missing prediction
    """

    indices = slice(90, None)

    # total count, datum 0, gt 0, pd 0, score 0, datum 1, gt 1, pd 1, score 1
    cm_gt0_pd0 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    cm_gt0_pd1 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    cm_gt1_pd0 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    cm_gt1_pd1 = np.array(
        [-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0]
    )
    expected_cm = np.array(
        [[cm_gt0_pd0, cm_gt0_pd1], [cm_gt1_pd0, cm_gt1_pd1]]
    )
    assert np.isclose(confusion_matrix[0, indices, :, :, :], expected_cm).all()

    # total count, datum 0, pd 0, score 0, datum 1, pd 1, score 1
    hal_pd0 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0])
    hal_pd1 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0])
    expected_hallucinations = np.array([hal_pd0, hal_pd1])
    assert np.isclose(
        hallucinations[0, indices, :, :], expected_hallucinations
    ).all()

    # total count, datum 0, gt 0, datum1, gt 1
    misprd_gt0 = np.array([4.0, 0.0, 0.0, 1.0, 2.0])
    misprd_gt1 = np.array([1.0, 1.0, 1.0, -1.0, -1.0])
    expected_missing_predictions = np.array([misprd_gt0, misprd_gt1])
    assert np.isclose(
        missing_predictions[0, indices, :, :], expected_missing_predictions
    ).all()


def _filter_out_zero_counts(cm: dict, hl: dict, mp: dict):
    gt_labels = list(mp.keys())
    pd_labels = list(hl.keys())

    for gt_label in gt_labels:
        for pd_label in pd_labels:
            if cm[gt_label][pd_label]["count"] == 0:
                cm[gt_label].pop(pd_label)
        if len(cm[gt_label]) == 0:
            cm.pop(gt_label)

    for pd_label in pd_labels:
        if hl[pd_label]["count"] == 0:
            hl.pop(pd_label)

    for gt_label in gt_labels:
        if mp[gt_label]["count"] == 0:
            mp.pop(gt_label)


def test_confusion_matrix(
    detections_for_detailed_counting: list[Detection],
    rect1: tuple[float, float, float, float],
    rect2: tuple[float, float, float, float],
    rect3: tuple[float, float, float, float],
    rect4: tuple[float, float, float, float],
    rect5: tuple[float, float, float, float],
):
    loader = DataLoader()
    loader.add_bounding_boxes(detections_for_detailed_counting)
    evaluator = loader.finalize()

    assert evaluator.ignored_prediction_labels == [
        "not_v2",
        "hallucination",
    ]
    assert evaluator.missing_prediction_labels == [
        "missed_detection",
        "v2",
    ]
    assert evaluator.n_datums == 2
    assert evaluator.n_labels == 6
    assert evaluator.n_groundtruths == 4
    assert evaluator.n_predictions == 4

    actual_metrics = evaluator.compute_confusion_matrix(
        iou_thresholds=[0.5],
        score_thresholds=[0.05, 0.3, 0.35, 0.45, 0.55, 0.95],
        number_of_examples=1,
    )

    rect1_dict = _convert_example_to_dict(np.array(rect1))
    rect2_dict = _convert_example_to_dict(np.array(rect2))
    rect3_dict = _convert_example_to_dict(np.array(rect3))
    rect4_dict = _convert_example_to_dict(np.array(rect4))
    rect5_dict = _convert_example_to_dict(np.array(rect5))

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect1_dict,
                                    "prediction": rect1_dict,
                                    "score": 0.5,
                                }
                            ],
                        }
                    }
                },
                "hallucinations": {
                    "not_v2": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid1",
                                "prediction": rect5_dict,
                                "score": 0.30000001192092896,
                            }
                        ],
                    },
                    "hallucination": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid1",
                                "prediction": rect4_dict,
                                "score": 0.10000000149011612,
                            }
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": rect2_dict,
                                "score": 0.5,
                            }
                        ],
                    },
                },
                "missing_predictions": {
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid2", "groundtruth": rect1_dict}
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.05,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect1_dict,
                                    "prediction": rect1_dict,
                                    "score": 0.5,
                                }
                            ],
                        }
                    }
                },
                "hallucinations": {
                    "not_v2": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid1",
                                "prediction": rect5_dict,
                                "score": 0.30000001192092896,
                            }
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": rect2_dict,
                                "score": 0.5,
                            }
                        ],
                    },
                },
                "missing_predictions": {
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid2", "groundtruth": rect1_dict}
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.3,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect1_dict,
                                    "prediction": rect1_dict,
                                    "score": 0.5,
                                }
                            ],
                        }
                    }
                },
                "hallucinations": {
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": rect2_dict,
                                "score": 0.5,
                            }
                        ],
                    }
                },
                "missing_predictions": {
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid2", "groundtruth": rect1_dict}
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.35,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect1_dict,
                                    "prediction": rect1_dict,
                                    "score": 0.5,
                                }
                            ],
                        }
                    }
                },
                "hallucinations": {
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": rect2_dict,
                                "score": 0.5,
                            }
                        ],
                    }
                },
                "missing_predictions": {
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid2", "groundtruth": rect1_dict}
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.45,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {},
                "hallucinations": {},
                "missing_predictions": {
                    "v1": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect1_dict}
                        ],
                    },
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid2", "groundtruth": rect1_dict}
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.55,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {},
                "hallucinations": {},
                "missing_predictions": {
                    "v1": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect1_dict}
                        ],
                    },
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid2", "groundtruth": rect1_dict}
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.95,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 1,
            },
        },
    ]
    for m in actual_metrics:
        _filter_out_zero_counts(
            m["value"]["confusion_matrix"],
            m["value"]["hallucinations"],
            m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics

    # test at lower IOU threshold

    actual_metrics = evaluator.compute_confusion_matrix(
        iou_thresholds=[0.45],
        score_thresholds=[0.05, 0.3, 0.35, 0.45, 0.55, 0.95],
        number_of_examples=1,
    )

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect1_dict,
                                    "prediction": rect1_dict,
                                    "score": 0.5,
                                }
                            ],
                        }
                    },
                    "v2": {
                        "not_v2": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect3_dict,
                                    "prediction": rect5_dict,
                                    "score": 0.30000001192092896,
                                }
                            ],
                        }
                    },
                },
                "hallucinations": {
                    "hallucination": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid1",
                                "prediction": rect4_dict,
                                "score": 0.10000000149011612,
                            }
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": rect2_dict,
                                "score": 0.5,
                            }
                        ],
                    },
                },
                "missing_predictions": {
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "groundtruth": rect1_dict,
                            }
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.05,
                "iou_threshold": 0.45,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect1_dict,
                                    "prediction": rect1_dict,
                                    "score": 0.5,
                                }
                            ],
                        }
                    },
                    "v2": {
                        "not_v2": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect3_dict,
                                    "prediction": rect5_dict,
                                    "score": 0.30000001192092896,
                                }
                            ],
                        }
                    },
                },
                "hallucinations": {
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": rect2_dict,
                                "score": 0.5,
                            }
                        ],
                    }
                },
                "missing_predictions": {
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "groundtruth": rect1_dict,
                            }
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.3,
                "iou_threshold": 0.45,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect1_dict,
                                    "prediction": rect1_dict,
                                    "score": 0.5,
                                }
                            ],
                        }
                    }
                },
                "hallucinations": {
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": rect2_dict,
                                "score": 0.5,
                            }
                        ],
                    }
                },
                "missing_predictions": {
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "groundtruth": rect1_dict,
                            }
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.35,
                "iou_threshold": 0.45,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": rect1_dict,
                                    "prediction": rect1_dict,
                                    "score": 0.5,
                                }
                            ],
                        }
                    }
                },
                "hallucinations": {
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": rect2_dict,
                                "score": 0.5,
                            }
                        ],
                    }
                },
                "missing_predictions": {
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "groundtruth": rect1_dict,
                            }
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.45,
                "iou_threshold": 0.45,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {},
                "hallucinations": {},
                "missing_predictions": {
                    "v1": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid1",
                                "groundtruth": rect1_dict,
                            }
                        ],
                    },
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "groundtruth": rect1_dict,
                            }
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.55,
                "iou_threshold": 0.45,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {},
                "hallucinations": {},
                "missing_predictions": {
                    "v1": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid1",
                                "groundtruth": rect1_dict,
                            }
                        ],
                    },
                    "missed_detection": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect2_dict}
                        ],
                    },
                    "v2": {
                        "count": 1,
                        "examples": [
                            {"datum": "uid1", "groundtruth": rect3_dict}
                        ],
                    },
                    "low_iou": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "groundtruth": rect1_dict,
                            }
                        ],
                    },
                },
            },
            "parameters": {
                "score_threshold": 0.95,
                "iou_threshold": 0.45,
                "maximum_number_of_examples": 1,
            },
        },
    ]

    for m in actual_metrics:
        _filter_out_zero_counts(
            m["value"]["confusion_matrix"],
            m["value"]["hallucinations"],
            m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_using_torch_metrics_example(
    torchmetrics_detections: list[Detection],
):
    """
    cf with torch metrics/pycocotools results listed here:
    https://github.com/Lightning-AI/metrics/blob/107dbfd5fb158b7ae6d76281df44bd94c836bfce/tests/unittests/detection/test_map.py#L231
    """
    loader = DataLoader()
    loader.add_bounding_boxes(torchmetrics_detections)
    evaluator = loader.finalize()

    assert evaluator.ignored_prediction_labels == ["3"]
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 4
    assert evaluator.n_labels == 6
    assert evaluator.n_groundtruths == 20
    assert evaluator.n_predictions == 19

    actual_metrics = evaluator.compute_confusion_matrix(
        iou_thresholds=[0.5, 0.9],
        score_thresholds=[0.05, 0.25, 0.35, 0.55, 0.75, 0.8, 0.85, 0.95],
        number_of_examples=0,
    )

    assert len(actual_metrics) == 16

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "4": {"4": {"count": 2, "examples": []}},
                    "2": {
                        "2": {"count": 1, "examples": []},
                        "3": {"count": 1, "examples": []},
                    },
                    "1": {"1": {"count": 1, "examples": []}},
                    "0": {"0": {"count": 5, "examples": []}},
                    "49": {"49": {"count": 9, "examples": []}},
                },
                "hallucinations": {},
                "missing_predictions": {"49": {"count": 1, "examples": []}},
            },
            "parameters": {
                "score_threshold": 0.05,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "4": {"4": {"count": 1, "examples": []}},
                    "2": {
                        "2": {"count": 1, "examples": []},
                        "3": {"count": 1, "examples": []},
                    },
                    "1": {"1": {"count": 1, "examples": []}},
                    "0": {"0": {"count": 5, "examples": []}},
                    "49": {"49": {"count": 6, "examples": []}},
                },
                "hallucinations": {},
                "missing_predictions": {
                    "4": {"count": 1, "examples": []},
                    "49": {"count": 4, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.25,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "4": {"4": {"count": 1, "examples": []}},
                    "2": {"2": {"count": 1, "examples": []}},
                    "0": {"0": {"count": 4, "examples": []}},
                    "49": {"49": {"count": 4, "examples": []}},
                },
                "hallucinations": {},
                "missing_predictions": {
                    "4": {"count": 1, "examples": []},
                    "2": {"count": 1, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 1, "examples": []},
                    "49": {"count": 6, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.35,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "2": {"2": {"count": 1, "examples": []}},
                    "0": {"0": {"count": 3, "examples": []}},
                    "49": {"49": {"count": 3, "examples": []}},
                },
                "hallucinations": {},
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 1, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 2, "examples": []},
                    "49": {"count": 7, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.55,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "0": {"0": {"count": 2, "examples": []}},
                    "49": {"49": {"count": 2, "examples": []}},
                },
                "hallucinations": {},
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 2, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 3, "examples": []},
                    "49": {"count": 8, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.75,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "0": {"0": {"count": 2, "examples": []}},
                    "49": {"49": {"count": 1, "examples": []}},
                },
                "hallucinations": {},
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 2, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 3, "examples": []},
                    "49": {"count": 9, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.8,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "0": {"0": {"count": 1, "examples": []}},
                    "49": {"49": {"count": 1, "examples": []}},
                },
                "hallucinations": {},
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 2, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 4, "examples": []},
                    "49": {"count": 9, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.85,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {"0": {"0": {"count": 1, "examples": []}}},
                "hallucinations": {},
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 2, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 4, "examples": []},
                    "49": {"count": 10, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.95,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "2": {"2": {"count": 1, "examples": []}},
                    "0": {"0": {"count": 1, "examples": []}},
                    "49": {"49": {"count": 2, "examples": []}},
                },
                "hallucinations": {
                    "4": {"count": 2, "examples": []},
                    "3": {"count": 1, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 4, "examples": []},
                    "49": {"count": 7, "examples": []},
                },
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 1, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 4, "examples": []},
                    "49": {"count": 8, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.05,
                "iou_threshold": 0.9,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "2": {"2": {"count": 1, "examples": []}},
                    "0": {"0": {"count": 1, "examples": []}},
                    "49": {"49": {"count": 2, "examples": []}},
                },
                "hallucinations": {
                    "4": {"count": 1, "examples": []},
                    "3": {"count": 1, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 4, "examples": []},
                    "49": {"count": 4, "examples": []},
                },
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 1, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 4, "examples": []},
                    "49": {"count": 8, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.25,
                "iou_threshold": 0.9,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "2": {"2": {"count": 1, "examples": []}},
                    "49": {"49": {"count": 2, "examples": []}},
                },
                "hallucinations": {
                    "4": {"count": 1, "examples": []},
                    "0": {"count": 4, "examples": []},
                    "49": {"count": 2, "examples": []},
                },
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 1, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 5, "examples": []},
                    "49": {"count": 8, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.35,
                "iou_threshold": 0.9,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "2": {"2": {"count": 1, "examples": []}},
                    "49": {"49": {"count": 1, "examples": []}},
                },
                "hallucinations": {
                    "0": {"count": 3, "examples": []},
                    "49": {"count": 2, "examples": []},
                },
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 1, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 5, "examples": []},
                    "49": {"count": 9, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.55,
                "iou_threshold": 0.9,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "49": {"49": {"count": 1, "examples": []}}
                },
                "hallucinations": {
                    "0": {"count": 2, "examples": []},
                    "49": {"count": 1, "examples": []},
                },
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 2, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 5, "examples": []},
                    "49": {"count": 9, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.75,
                "iou_threshold": 0.9,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "49": {"49": {"count": 1, "examples": []}}
                },
                "hallucinations": {"0": {"count": 2, "examples": []}},
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 2, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 5, "examples": []},
                    "49": {"count": 9, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.8,
                "iou_threshold": 0.9,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "49": {"49": {"count": 1, "examples": []}}
                },
                "hallucinations": {"0": {"count": 1, "examples": []}},
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 2, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 5, "examples": []},
                    "49": {"count": 9, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.85,
                "iou_threshold": 0.9,
                "maximum_number_of_examples": 0,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {},
                "hallucinations": {"0": {"count": 1, "examples": []}},
                "missing_predictions": {
                    "4": {"count": 2, "examples": []},
                    "2": {"count": 2, "examples": []},
                    "1": {"count": 1, "examples": []},
                    "0": {"count": 5, "examples": []},
                    "49": {"count": 10, "examples": []},
                },
            },
            "parameters": {
                "score_threshold": 0.95,
                "iou_threshold": 0.9,
                "maximum_number_of_examples": 0,
            },
        },
    ]
    for m in actual_metrics:
        _filter_out_zero_counts(
            m["value"]["confusion_matrix"],
            m["value"]["hallucinations"],
            m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_fp_hallucination_edge_case(
    detections_fp_hallucination_edge_case: list[Detection],
):

    loader = DataLoader()
    loader.add_bounding_boxes(detections_fp_hallucination_edge_case)
    evaluator = loader.finalize()

    assert evaluator.ignored_prediction_labels == []
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 2
    assert evaluator.n_labels == 1
    assert evaluator.n_groundtruths == 2
    assert evaluator.n_predictions == 2

    actual_metrics = evaluator.compute_confusion_matrix(
        iou_thresholds=[0.5],
        score_thresholds=[0.5, 0.85],
        number_of_examples=1,
    )

    assert len(actual_metrics) == 2

    actual_metrics = [m.to_dict() for m in actual_metrics]
    expected_metrics = [
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {
                    "v1": {
                        "v1": {
                            "count": 1,
                            "examples": [
                                {
                                    "datum": "uid1",
                                    "groundtruth": {
                                        "xmin": 0.0,
                                        "xmax": 5.0,
                                        "ymin": 0.0,
                                        "ymax": 5.0,
                                    },
                                    "prediction": {
                                        "xmin": 0.0,
                                        "xmax": 5.0,
                                        "ymin": 0.0,
                                        "ymax": 5.0,
                                    },
                                    "score": 0.800000011920929,
                                }
                            ],
                        }
                    }
                },
                "hallucinations": {
                    "v1": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "prediction": {
                                    "xmin": 10.0,
                                    "xmax": 20.0,
                                    "ymin": 10.0,
                                    "ymax": 20.0,
                                },
                                "score": 0.800000011920929,
                            }
                        ],
                    }
                },
                "missing_predictions": {
                    "v1": {
                        "count": 1,
                        "examples": [
                            {
                                "datum": "uid2",
                                "groundtruth": {
                                    "xmin": 0.0,
                                    "xmax": 5.0,
                                    "ymin": 0.0,
                                    "ymax": 5.0,
                                },
                            }
                        ],
                    }
                },
            },
            "parameters": {
                "score_threshold": 0.5,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 1,
            },
        },
        {
            "type": "ConfusionMatrix",
            "value": {
                "confusion_matrix": {},
                "hallucinations": {},
                "missing_predictions": {
                    "v1": {
                        "count": 2,
                        "examples": [
                            {
                                "datum": "uid1",
                                "groundtruth": {
                                    "xmin": 0.0,
                                    "xmax": 5.0,
                                    "ymin": 0.0,
                                    "ymax": 5.0,
                                },
                            }
                        ],
                    }
                },
            },
            "parameters": {
                "score_threshold": 0.85,
                "iou_threshold": 0.5,
                "maximum_number_of_examples": 1,
            },
        },
    ]
    for m in actual_metrics:
        _filter_out_zero_counts(
            m["value"]["confusion_matrix"],
            m["value"]["hallucinations"],
            m["value"]["missing_predictions"],
        )
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_confusion_matrix_ranked_pair_ordering(
    detection_ranked_pair_ordering: Detection,
    detection_ranked_pair_ordering_with_bitmasks: Detection,
    detection_ranked_pair_ordering_with_polygons: Detection,
):

    for input_, method in [
        (detection_ranked_pair_ordering, DataLoader.add_bounding_boxes),
        (
            detection_ranked_pair_ordering_with_bitmasks,
            DataLoader.add_bitmasks,
        ),
        (
            detection_ranked_pair_ordering_with_polygons,
            DataLoader.add_polygons,
        ),
    ]:
        loader = DataLoader()
        method(loader, detections=[input_])

        evaluator = loader.finalize()

        assert evaluator.metadata == {
            "ignored_prediction_labels": [
                "label4",
            ],
            "missing_prediction_labels": [],
            "n_datums": 1,
            "n_groundtruths": 3,
            "n_labels": 4,
            "n_predictions": 4,
        }

        actual_metrics = evaluator.compute_confusion_matrix(
            iou_thresholds=[0.5],
            score_thresholds=[0.0],
            number_of_examples=0,
        )

        actual_metrics = [m.to_dict() for m in actual_metrics]
        expected_metrics = [
            {
                "type": "ConfusionMatrix",
                "value": {
                    "confusion_matrix": {
                        "label1": {"label1": {"count": 1, "examples": []}},
                        "label2": {"label2": {"count": 1, "examples": []}},
                    },
                    "hallucinations": {
                        "label3": {"count": 1, "examples": []},
                        "label4": {"count": 1, "examples": []},
                    },
                    "missing_predictions": {
                        "label3": {"count": 1, "examples": []}
                    },
                },
                "parameters": {
                    "score_threshold": 0.0,
                    "iou_threshold": 0.5,
                    "maximum_number_of_examples": 0,
                },
            },
        ]
        for m in actual_metrics:
            _filter_out_zero_counts(
                m["value"]["confusion_matrix"],
                m["value"]["hallucinations"],
                m["value"]["missing_predictions"],
            )
            assert m in expected_metrics
        for m in expected_metrics:
            assert m in actual_metrics



from .annotation import Classification
from .computation import (
    compute_confusion_matrix,
    compute_precision_recall_rocauc,
)
from .manager import DataLoader, Evaluator
from .metric import Metric, MetricType

__all__ = [
    "Classification",
    "compute_precision_recall_rocauc",
    "compute_confusion_matrix",
    "MetricType",
    "DataLoader",
    "Evaluator",
    "Metric",
]



import json

import pytest
from valor_lite.text_generation import Context, Evaluator, QueryResponse
from valor_lite.text_generation.computation import calculate_answer_correctness


def test_answer_correctness_no_context(mock_client):

    evaluator = Evaluator(client=mock_client)
    with pytest.raises(ValueError):
        assert evaluator.compute_answer_correctness(
            response=QueryResponse(
                query="a",
                response="a",
            )
        )


def test_calculate_answer_correctness(mock_client):

    evaluator = Evaluator(client=mock_client)

    mock_client.returning = json.dumps(
        {
            "TP": ["a", "b", "c"],
            "FP": ["d", "e"],
            "FN": ["f"],
            "statements": ["v", "w", "x", "y", "z"],
        }
    )
    assert (
        calculate_answer_correctness(
            client=mock_client,
            system_prompt="",
            query="a",
            response="a",
            groundtruths=["a", "b", "c"],
        )
        == 2 / 3
    )
    assert evaluator.compute_answer_correctness(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(
                groundtruth=["a", "b", "c"],
            ),
        )
    ).to_dict() == {
        "type": "AnswerCorrectness",
        "value": 2 / 3,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = json.dumps(
        {"TP": [], "FP": ["a"], "FN": ["b"], "statements": ["c"]}
    )
    assert (
        calculate_answer_correctness(
            client=mock_client,
            system_prompt="",
            query="a",
            response="a",
            groundtruths=["a", "b", "c"],
        )
        == 0
    )
    assert evaluator.compute_answer_correctness(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(
                groundtruth=["a", "b", "c"],
            ),
        )
    ).to_dict() == {
        "type": "AnswerCorrectness",
        "value": 0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }



from dataclasses import dataclass, field


@dataclass
class Context:
    """
    Contextual ground truth and prediction.

    Attributes
    ----------
    groundtruth : list[str]
        The definitive context.
    prediction : list[str]
        Any retrieved context from a retrieval-augmented-generation (RAG) pipeline.

    Examples
    --------
    ... context = Context(
    ...     groundtruth=[...],
    ...     prediction=[...],
    ... )
    """

    groundtruth: list[str] = field(default_factory=list)
    prediction: list[str] = field(default_factory=list)


@dataclass
class QueryResponse:
    """
    Text generation data structure containing ground truths and predictions.

    Attributes
    ----------
    query : str
        The user query.
    response : str
        The language model's response.
    context : Context
        Any context provided to the model.

    Examples
    --------
    >>> query = QueryResponse(
    ...     query='When was George Washington born?',
    ...     response="February 22, 1732",
    ...     context=Context(
    ...         groundtruth=["02/22/1732"],
    ...         prediction=["02/22/1732"],
    ...     ),
    ... )
    """

    query: str
    response: str
    context: Context | None = field(default=None)



import numpy as np
import pytest
from shapely.geometry import Polygon as ShapelyPolygon
from valor_lite.object_detection import (
    Bitmask,
    BoundingBox,
    Detection,
    Polygon,
)


@pytest.fixture
def rect1() -> tuple[float, float, float, float]:
    """Box with area = 1500."""
    return (10.0, 60.0, 10.0, 40.0)


@pytest.fixture
def rect2() -> tuple[float, float, float, float]:
    """Box with area = 1100."""
    return (15.0, 70.0, 0.0, 20.0)


@pytest.fixture
def rect3() -> tuple[float, float, float, float]:
    """Box with area = 57,510."""
    return (87.0, 158.0, 10.0, 820.0)


@pytest.fixture
def rect4() -> tuple[float, float, float, float]:
    """Box with area = 90."""
    return (1.0, 10.0, 10.0, 20.0)


@pytest.fixture
def rect5() -> tuple[float, float, float, float]:
    """Box with partial overlap to rect3."""
    return (87, 158, 10, 400)


@pytest.fixture
def rect1_rotated_5_degrees_around_origin() -> list[tuple[float, float]]:
    """Box with area = 1500."""
    return [
        (9.090389553440874, 10.833504408394036),
        (58.90012445802815, 15.191291545776945),
        (56.28545217559841, 45.07713248852931),
        (6.475717271011129, 40.7193453511464),
        (9.090389553440874, 10.833504408394036),
    ]


@pytest.fixture
def rect2_rotated_5_degrees_around_origin() -> list[tuple[float, float]]:
    """Box with area = 1100."""
    return [
        (14.942920471376183, 1.3073361412148725),
        (69.7336288664222, 6.1009019923360714),
        (67.99051401146903, 26.024795954170983),
        (13.19980561642302, 21.231230103049782),
        (14.942920471376183, 1.3073361412148725),
    ]


@pytest.fixture
def rect3_rotated_5_degrees_around_origin() -> list[tuple[float, float]]:
    """Box with area = 57,510."""
    return [
        (85.79738130650527, 17.544496599963715),
        (156.52720487101922, 23.732554335047446),
        (85.9310532454161, 830.6502597893614),
        (15.20122968090216, 824.4622020542777),
        (85.79738130650527, 17.544496599963715),
    ]


@pytest.fixture
def basic_detections_first_class(
    rect1: tuple[float, float, float, float],
    rect2: tuple[float, float, float, float],
) -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["v1"],
                ),
            ],
            predictions=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["v1"],
                    scores=[0.3],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=rect2[0],
                    xmax=rect2[1],
                    ymin=rect2[2],
                    ymax=rect2[3],
                    labels=["v1"],
                ),
            ],
            predictions=[],
        ),
    ]


@pytest.fixture
def basic_detections_second_class(
    rect2: tuple[float, float, float, float],
    rect3: tuple[float, float, float, float],
) -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=rect3[0],
                    xmax=rect3[1],
                    ymin=rect3[2],
                    ymax=rect3[3],
                    labels=["v2"],
                ),
            ],
            predictions=[],
        ),
        Detection(
            uid="uid2",
            groundtruths=[],
            predictions=[
                BoundingBox(
                    xmin=rect2[0],
                    xmax=rect2[1],
                    ymin=rect2[2],
                    ymax=rect2[3],
                    labels=["v2"],
                    scores=[0.98],
                ),
            ],
        ),
    ]


@pytest.fixture
def basic_detections(
    rect1: tuple[float, float, float, float],
    rect2: tuple[float, float, float, float],
    rect3: tuple[float, float, float, float],
) -> list[Detection]:
    """Combines the labels from basic_detections_first_class and basic_detections_second_class."""
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["v1"],
                ),
                BoundingBox(
                    xmin=rect3[0],
                    xmax=rect3[1],
                    ymin=rect3[2],
                    ymax=rect3[3],
                    labels=["v2"],
                ),
            ],
            predictions=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["v1"],
                    scores=[0.3],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=rect2[0],
                    xmax=rect2[1],
                    ymin=rect2[2],
                    ymax=rect2[3],
                    labels=["v1"],
                ),
            ],
            predictions=[
                BoundingBox(
                    xmin=rect2[0],
                    xmax=rect2[1],
                    ymin=rect2[2],
                    ymax=rect2[3],
                    labels=["v2"],
                    scores=[0.98],
                ),
            ],
        ),
    ]


@pytest.fixture
def basic_rotated_detections_first_class(
    rect1_rotated_5_degrees_around_origin: tuple[float, float, float, float],
    rect2_rotated_5_degrees_around_origin: tuple[float, float, float, float],
) -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                Polygon(
                    shape=ShapelyPolygon(
                        rect1_rotated_5_degrees_around_origin
                    ),
                    labels=["v1"],
                ),
            ],
            predictions=[
                Polygon(
                    shape=ShapelyPolygon(
                        rect1_rotated_5_degrees_around_origin
                    ),
                    labels=["v1"],
                    scores=[0.3],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                Polygon(
                    shape=ShapelyPolygon(
                        rect2_rotated_5_degrees_around_origin
                    ),
                    labels=["v1"],
                ),
            ],
            predictions=[],
        ),
    ]


@pytest.fixture
def basic_rotated_detections_second_class(
    rect2_rotated_5_degrees_around_origin: tuple[float, float, float, float],
    rect3_rotated_5_degrees_around_origin: tuple[float, float, float, float],
) -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                Polygon(
                    shape=ShapelyPolygon(
                        rect3_rotated_5_degrees_around_origin
                    ),
                    labels=["v2"],
                ),
            ],
            predictions=[],
        ),
        Detection(
            uid="uid2",
            groundtruths=[],
            predictions=[
                Polygon(
                    shape=ShapelyPolygon(
                        rect2_rotated_5_degrees_around_origin
                    ),
                    labels=["v2"],
                    scores=[0.98],
                ),
            ],
        ),
    ]


@pytest.fixture
def torchmetrics_detections() -> list[Detection]:
    """Creates a model called "test_model" with some predicted
    detections on the dataset "test_dataset". These predictions are taken
    from a torchmetrics unit test (see test_metrics.py)
    """

    # predictions for four images taken from
    # https://github.com/Lightning-AI/metrics/blob/107dbfd5fb158b7ae6d76281df44bd94c836bfce/tests/unittests/detection/test_map.py#L59

    groundtruths = [
        {"boxes": [[214.1500, 41.2900, 562.4100, 285.0700]], "labels": ["4"]},
        {
            "boxes": [
                [13.00, 22.75, 548.98, 632.42],
                [1.66, 3.32, 270.26, 275.23],
            ],
            "labels": ["2", "2"],
        },
        {
            "boxes": [
                [61.87, 276.25, 358.29, 379.43],
                [2.75, 3.66, 162.15, 316.06],
                [295.55, 93.96, 313.97, 152.79],
                [326.94, 97.05, 340.49, 122.98],
                [356.62, 95.47, 372.33, 147.55],
                [462.08, 105.09, 493.74, 146.99],
                [277.11, 103.84, 292.44, 150.72],
            ],
            "labels": ["4", "1", "0", "0", "0", "0", "0"],
        },
        {
            "boxes": [
                [72.92, 45.96, 91.23, 80.57],
                [50.17, 45.34, 71.28, 79.83],
                [81.28, 47.04, 98.66, 78.50],
                [63.96, 46.17, 84.35, 80.48],
                [75.29, 23.01, 91.85, 50.85],
                [56.39, 21.65, 75.66, 45.54],
                [73.14, 1.10, 98.96, 28.33],
                [62.34, 55.23, 78.14, 79.57],
                [44.17, 45.78, 63.99, 78.48],
                [58.18, 44.80, 66.42, 56.25],
            ],
            "labels": [
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
                "49",
            ],
        },
    ]
    predictions = [
        {
            "boxes": [[258.15, 41.29, 606.41, 285.07]],
            "scores": [0.236],
            "labels": ["4"],
        },
        {
            "boxes": [
                [61.00, 22.75, 565.00, 632.42],
                [12.66, 3.32, 281.26, 275.23],
            ],
            "scores": [0.318, 0.726],
            "labels": ["3", "2"],
        },
        {
            "boxes": [
                [87.87, 276.25, 384.29, 379.43],
                [0.00, 3.66, 142.15, 316.06],
                [296.55, 93.96, 314.97, 152.79],
                [328.94, 97.05, 342.49, 122.98],
                [356.62, 95.47, 372.33, 147.55],
                [464.08, 105.09, 495.74, 146.99],
                [276.11, 103.84, 291.44, 150.72],
            ],
            "scores": [0.546, 0.3, 0.407, 0.611, 0.335, 0.805, 0.953],
            "labels": ["4", "1", "0", "0", "0", "0", "0"],
        },
        {
            "boxes": [
                [72.92, 45.96, 91.23, 80.57],
                [45.17, 45.34, 66.28, 79.83],
                [82.28, 47.04, 99.66, 78.50],
                [59.96, 46.17, 80.35, 80.48],
                [75.29, 23.01, 91.85, 50.85],
                [71.14, 1.10, 96.96, 28.33],
                [61.34, 55.23, 77.14, 79.57],
                [41.17, 45.78, 60.99, 78.48],
                [56.18, 44.80, 64.42, 56.25],
            ],
            "scores": [
                0.532,
                0.204,
                0.782,
                0.202,
                0.883,
                0.271,
                0.561,
                0.204,
                0.349,
            ],
            "labels": ["49", "49", "49", "49", "49", "49", "49", "49", "49"],
        },
    ]

    return [
        Detection(
            uid=str(idx),
            groundtruths=[
                BoundingBox(
                    xmin=box[0],
                    ymin=box[1],
                    xmax=box[2],
                    ymax=box[3],
                    labels=[label_value],
                )
                for box, label_value in zip(gt["boxes"], gt["labels"])
            ],
            predictions=[
                BoundingBox(
                    xmin=box[0],
                    ymin=box[1],
                    xmax=box[2],
                    ymax=box[3],
                    labels=[label_value],
                    scores=[score],
                )
                for box, label_value, score in zip(
                    pd["boxes"], pd["labels"], pd["scores"]
                )
            ],
        )
        for idx, (gt, pd) in enumerate(zip(groundtruths, predictions))
    ]


@pytest.fixture
def false_negatives_single_datum_baseline_detections() -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.8],
                ),
                BoundingBox(
                    xmin=100,
                    xmax=110,
                    ymin=100,
                    ymax=200,
                    labels=["value"],
                    scores=[0.7],
                ),
            ],
        )
    ]


@pytest.fixture
def false_negatives_single_datum_detections() -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.8],
                ),
                BoundingBox(
                    xmin=100,
                    xmax=110,
                    ymin=100,
                    ymax=200,
                    labels=["value"],
                    scores=[0.9],
                ),
            ],
        )
    ]


@pytest.fixture
def false_negatives_two_datums_one_empty_low_confidence_of_fp_detections() -> (
    list[Detection]
):

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.8],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.7],
                ),
            ],
        ),
    ]


@pytest.fixture
def false_negatives_two_datums_one_empty_high_confidence_of_fp_detections() -> (
    list[Detection]
):

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.8],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.9],
                ),
            ],
        ),
    ]


@pytest.fixture
def false_negatives_two_datums_one_only_with_different_class_low_confidence_of_fp_detections() -> (
    list[Detection]
):

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.8],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["other value"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.7],
                ),
            ],
        ),
    ]


@pytest.fixture
def false_negatives_two_images_one_only_with_different_class_high_confidence_of_fp_detections() -> (
    list[Detection]
):

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.8],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["other value"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["value"],
                    scores=[0.9],
                ),
            ],
        ),
    ]


@pytest.fixture
def detections_fp_hallucination_edge_case() -> list[Detection]:
    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=0,
                    xmax=5,
                    ymin=0,
                    ymax=5,
                    labels=["v1"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=0,
                    xmax=5,
                    ymin=0,
                    ymax=5,
                    labels=["v1"],
                    scores=[0.8],
                )
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=0,
                    xmax=5,
                    ymin=0,
                    ymax=5,
                    labels=["v1"],
                )
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["v1"],
                    scores=[0.8],
                )
            ],
        ),
    ]


@pytest.fixture
def detections_tp_deassignment_edge_case() -> list[Detection]:
    return [
        Detection(
            uid="uid0",
            groundtruths=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["v1"],
                ),
                BoundingBox(
                    xmin=10,
                    xmax=15,
                    ymin=20,
                    ymax=25,
                    labels=["v1"],
                ),
            ],
            predictions=[
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=10,
                    ymax=20,
                    labels=["v1"],
                    scores=[0.78],
                ),
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=12,
                    ymax=22,
                    labels=["v1"],
                    scores=[0.96],
                ),
                BoundingBox(
                    xmin=10,
                    xmax=20,
                    ymin=12,
                    ymax=22,
                    labels=["v1"],
                    scores=[0.96],
                ),
                BoundingBox(
                    xmin=101,
                    xmax=102,
                    ymin=101,
                    ymax=102,
                    labels=["v1"],
                    scores=[0.87],
                ),
            ],
        ),
    ]


@pytest.fixture
def detection_ranked_pair_ordering() -> Detection:

    gts = {
        "boxes": [
            (2, 10, 2, 10),
            (2, 10, 2, 10),
            (2, 10, 2, 10),
        ],
        "label_values": ["label1", "label2", "label3"],
    }

    # labels 1 and 2 have IOU==1, labels 3 and 4 have IOU==0
    preds = {
        "boxes": [
            (2, 10, 2, 10),
            (2, 10, 2, 10),
            (0, 1, 0, 1),
            (0, 1, 0, 1),
        ],
        "label_values": ["label1", "label2", "label3", "label4"],
        "scores": [
            0.3,
            0.93,
            0.92,
            0.94,
        ],
    }

    groundtruths = [
        BoundingBox(
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            labels=[label_value],
        )
        for (xmin, xmax, ymin, ymax), label_value in zip(
            gts["boxes"], gts["label_values"]
        )
    ]

    predictions = [
        BoundingBox(
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            labels=[label_value],
            scores=[score],
        )
        for (xmin, xmax, ymin, ymax), label_value, score in zip(
            preds["boxes"], preds["label_values"], preds["scores"]
        )
    ]

    return Detection(
        uid="uid1", groundtruths=groundtruths, predictions=predictions
    )


@pytest.fixture
def detection_ranked_pair_ordering_with_bitmasks() -> Detection:

    bitmask1 = np.zeros((100, 50), dtype=np.bool_)
    bitmask1[79, 31] = True

    bitmask2 = np.zeros((100, 50), dtype=np.bool_)
    bitmask2[80:, 32:] = True

    gts = {
        "bitmasks": [
            bitmask1,
            bitmask1,
            bitmask1,
        ],
        "label_values": ["label1", "label2", "label3"],
    }

    # labels 1 and 2 have IOU==1, labels 3 and 4 have IOU==0
    preds = {
        "bitmasks": [bitmask1, bitmask1, bitmask2, bitmask2],
        "label_values": ["label1", "label2", "label3", "label4"],
        "scores": [
            0.3,
            0.93,
            0.92,
            0.94,
        ],
    }

    groundtruths = [
        Bitmask(
            mask=mask,
            labels=[label_value],
        )
        for mask, label_value in zip(gts["bitmasks"], gts["label_values"])
    ]

    predictions = [
        Bitmask(
            mask=mask,
            labels=[label_value],
            scores=[score],
        )
        for mask, label_value, score in zip(
            preds["bitmasks"], preds["label_values"], preds["scores"]
        )
    ]

    return Detection(
        uid="uid1", groundtruths=groundtruths, predictions=predictions
    )


@pytest.fixture
def detection_ranked_pair_ordering_with_polygons(
    rect1_rotated_5_degrees_around_origin: list[tuple[float, float]],
    rect3_rotated_5_degrees_around_origin: list[tuple[float, float]],
) -> Detection:
    gts = {
        "polygons": [
            rect1_rotated_5_degrees_around_origin,
            rect1_rotated_5_degrees_around_origin,
            rect1_rotated_5_degrees_around_origin,
        ],
        "label_values": ["label1", "label2", "label3"],
    }

    # labels 1 and 2 have IOU==1, labels 3 and 4 have IOU==0
    preds = {
        "polygons": [
            rect1_rotated_5_degrees_around_origin,
            rect1_rotated_5_degrees_around_origin,
            rect3_rotated_5_degrees_around_origin,
            rect3_rotated_5_degrees_around_origin,
        ],
        "label_values": ["label1", "label2", "label3", "label4"],
        "scores": [
            0.3,
            0.93,
            0.92,
            0.94,
        ],
    }

    groundtruths = [
        Polygon(
            shape=ShapelyPolygon(polygon),
            labels=[label_value],
        )
        for polygon, label_value in zip(gts["polygons"], gts["label_values"])
    ]

    predictions = [
        Polygon(
            shape=ShapelyPolygon(polygon),
            labels=[label_value],
            scores=[score],
        )
        for polygon, label_value, score in zip(
            preds["polygons"], preds["label_values"], preds["scores"]
        )
    ]

    return Detection(
        uid="uid1", groundtruths=groundtruths, predictions=predictions
    )


@pytest.fixture
def detections_no_groundtruths() -> list[Detection]:
    return [
        Detection(
            uid="uid",
            groundtruths=[],
            predictions=[
                BoundingBox(
                    xmin=0,
                    xmax=10,
                    ymin=0,
                    ymax=10,
                    labels=["v1"],
                    scores=[1.0],
                ),
            ],
        ),
        Detection(
            uid="uid",
            groundtruths=[],
            predictions=[
                BoundingBox(
                    xmin=0,
                    xmax=10,
                    ymin=0,
                    ymax=10,
                    labels=["v1"],
                    scores=[1.0],
                ),
            ],
        ),
    ]


@pytest.fixture
def detections_no_predictions() -> list[Detection]:
    return [
        Detection(
            uid="uid",
            groundtruths=[
                BoundingBox(xmin=0, xmax=10, ymin=0, ymax=10, labels=["v1"]),
            ],
            predictions=[],
        ),
        Detection(
            uid="uid",
            groundtruths=[
                BoundingBox(xmin=0, xmax=10, ymin=0, ymax=10, labels=["v1"]),
            ],
            predictions=[],
        ),
    ]


@pytest.fixture
def detections_for_detailed_counting(
    rect1: tuple[float, float, float, float],
    rect2: tuple[float, float, float, float],
    rect3: tuple[float, float, float, float],
    rect4: tuple[float, float, float, float],
    rect5: tuple[float, float, float, float],
) -> list[Detection]:

    return [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["v1"],
                ),
                BoundingBox(
                    xmin=rect2[0],
                    xmax=rect2[1],
                    ymin=rect2[2],
                    ymax=rect2[3],
                    labels=["missed_detection"],
                ),
                BoundingBox(
                    xmin=rect3[0],
                    xmax=rect3[1],
                    ymin=rect3[2],
                    ymax=rect3[3],
                    labels=["v2"],
                ),
            ],
            predictions=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["v1"],
                    scores=[0.5],
                ),
                BoundingBox(
                    xmin=rect5[0],
                    xmax=rect5[1],
                    ymin=rect5[2],
                    ymax=rect5[3],
                    labels=["not_v2"],
                    scores=[0.3],
                ),
                BoundingBox(
                    xmin=rect4[0],
                    xmax=rect4[1],
                    ymin=rect4[2],
                    ymax=rect4[3],
                    labels=["hallucination"],
                    scores=[0.1],
                ),
            ],
        ),
        Detection(
            uid="uid2",
            groundtruths=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["low_iou"],
                ),
            ],
            predictions=[
                BoundingBox(
                    xmin=rect2[0],
                    xmax=rect2[1],
                    ymin=rect2[2],
                    ymax=rect2[3],
                    labels=["low_iou"],
                    scores=[0.5],
                ),
            ],
        ),
    ]



import numpy as np
import pytest
from shapely.geometry import Polygon as ShapelyPolygon
from valor_lite.object_detection import (
    Bitmask,
    BoundingBox,
    DataLoader,
    Detection,
    Polygon,
)


def test_no_data():
    loader = DataLoader()
    with pytest.raises(ValueError):
        loader.finalize()


def test_iou_computation():

    detection = Detection(
        uid="uid",
        groundtruths=[
            BoundingBox(xmin=0, xmax=10, ymin=0, ymax=10, labels=["0"]),
            BoundingBox(xmin=100, xmax=110, ymin=100, ymax=110, labels=["0"]),
            BoundingBox(
                xmin=1000, xmax=1100, ymin=1000, ymax=1100, labels=["0"]
            ),
        ],
        predictions=[
            BoundingBox(
                xmin=1,
                xmax=11,
                ymin=1,
                ymax=11,
                labels=["0", "1", "2"],
                scores=[0.5, 0.25, 0.25],
            ),
            BoundingBox(
                xmin=105,
                xmax=116,
                ymin=105,
                ymax=116,
                labels=["0", "1", "2"],
                scores=[0.5, 0.25, 0.25],
            ),
        ],
    )

    loader = DataLoader()
    loader.add_bounding_boxes([detection])

    assert len(loader.pairs) == 1

    # show that three unique IOUs exist
    unique_ious = np.unique(loader.pairs[0][:, 3])
    assert np.isclose(
        unique_ious, np.array([0.0, 0.12755102, 0.68067227])
    ).all()


def test_mixed_annotations(
    rect1: tuple[float, float, float, float],
    rect1_rotated_5_degrees_around_origin: tuple[float, float, float, float],
):
    """Check that we throw an error if the user tries to mix annotation types."""

    # test add_bounding_box
    mixed_detections = [
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["v1"],
                ),
            ],
            predictions=[
                Polygon(
                    shape=ShapelyPolygon(
                        rect1_rotated_5_degrees_around_origin
                    ),
                    labels=["v1"],
                    scores=[0.3],
                ),
            ],
        ),
        Detection(
            uid="uid1",
            groundtruths=[
                BoundingBox(
                    xmin=rect1[0],
                    xmax=rect1[1],
                    ymin=rect1[2],
                    ymax=rect1[3],
                    labels=["v1"],
                ),
            ],
            predictions=[
                Bitmask(
                    mask=np.ones((80, 32), dtype=bool),
                    labels=["v1"],
                    scores=[0.3],
                ),
            ],
        ),
        Detection(
            uid="uid1",
            groundtruths=[
                Bitmask(
                    mask=np.ones((80, 32), dtype=bool),
                    labels=["v1"],
                    scores=[0.3],
                ),
            ],
            predictions=[
                Polygon(
                    shape=ShapelyPolygon(
                        rect1_rotated_5_degrees_around_origin
                    ),
                    labels=["v1"],
                    scores=[0.3],
                ),
            ],
        ),
    ]

    loader = DataLoader()

    for detection in mixed_detections:

        # anything can be converted to a bbox
        loader.add_bounding_boxes([detection])

        with pytest.raises(AttributeError) as e:
            loader.add_polygons([detection])
        assert "no attribute 'shape'" in str(e)

        with pytest.raises(AttributeError) as e:
            loader.add_bitmasks([detection])
        assert "no attribute 'mask'" in str(e)



from dataclasses import dataclass
from enum import Enum

from valor_lite.schemas import BaseMetric


class MetricType(str, Enum):
    Counts = "Counts"
    Accuracy = "Accuracy"
    Precision = "Precision"
    Recall = "Recall"
    F1 = "F1"
    AP = "AP"
    AR = "AR"
    mAP = "mAP"
    mAR = "mAR"
    APAveragedOverIOUs = "APAveragedOverIOUs"
    mAPAveragedOverIOUs = "mAPAveragedOverIOUs"
    ARAveragedOverScores = "ARAveragedOverScores"
    mARAveragedOverScores = "mARAveragedOverScores"
    PrecisionRecallCurve = "PrecisionRecallCurve"
    ConfusionMatrix = "ConfusionMatrix"


@dataclass
class Metric(BaseMetric):
    """
    Object Detection Metric.

    Attributes
    ----------
    type : str
        The metric type.
    value : int | float | dict
        The metric value.
    parameters : dict[str, Any]
        A dictionary containing metric parameters.
    """

    def __post_init__(self):
        if not isinstance(self.type, str):
            raise TypeError(
                f"Metric type should be of type 'str': {self.type}"
            )
        elif not isinstance(self.value, (int, float, dict)):
            raise TypeError(
                f"Metric value must be of type 'int', 'float' or 'dict': {self.value}"
            )
        elif not isinstance(self.parameters, dict):
            raise TypeError(
                f"Metric parameters must be of type 'dict[str, Any]': {self.parameters}"
            )
        elif not all([isinstance(k, str) for k in self.parameters.keys()]):
            raise TypeError(
                f"Metric parameter dictionary should only have keys with type 'str': {self.parameters}"
            )

    @classmethod
    def precision(
        cls,
        value: float,
        label: str,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        Precision metric for a specific class label in object detection.

        This class encapsulates a metric value for a particular class label,
        along with the associated Intersection over Union (IOU) threshold and
        confidence score threshold.

        Parameters
        ----------
        value : float
            The metric value.
        label : str
            The class label for which the metric is calculated.
        iou_threshold : float
            The IOU threshold used to determine matches between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Precision.value,
            value=value,
            parameters={
                "label": label,
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def recall(
        cls,
        value: float,
        label: str,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        Recall metric for a specific class label in object detection.

        This class encapsulates a metric value for a particular class label,
        along with the associated Intersection over Union (IOU) threshold and
        confidence score threshold.

        Parameters
        ----------
        value : float
            The metric value.
        label : str
            The class label for which the metric is calculated.
        iou_threshold : float
            The IOU threshold used to determine matches between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Recall.value,
            value=value,
            parameters={
                "label": label,
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def f1_score(
        cls,
        value: float,
        label: str,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        F1 score for a specific class label in object detection.

        This class encapsulates a metric value for a particular class label,
        along with the associated Intersection over Union (IOU) threshold and
        confidence score threshold.

        Parameters
        ----------
        value : float
            The metric value.
        label : str
            The class label for which the metric is calculated.
        iou_threshold : float
            The IOU threshold used to determine matches between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.F1.value,
            value=value,
            parameters={
                "label": label,
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def accuracy(
        cls,
        value: float,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        Accuracy metric for the object detection task type.

        This class encapsulates a metric value at a specific Intersection
        over Union (IOU) threshold and confidence score threshold.

        Parameters
        ----------
        value : float
            The metric value.
        iou_threshold : float
            The IOU threshold used to determine matches between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Accuracy.value,
            value=value,
            parameters={
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def average_precision(
        cls,
        value: float,
        iou_threshold: float,
        label: str,
    ):
        """
        Average Precision (AP) metric for object detection tasks.

        The AP computation uses 101-point interpolation, which calculates the average
        precision by interpolating the precision-recall curve at 101 evenly spaced recall
        levels from 0 to 1.

        Parameters
        ----------
        value : float
            The average precision value.
        iou_threshold : float
            The IOU threshold used to compute the AP.
        label : str
            The class label for which the AP is computed.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.AP.value,
            value=value,
            parameters={
                "iou_threshold": iou_threshold,
                "label": label,
            },
        )

    @classmethod
    def mean_average_precision(
        cls,
        value: float,
        iou_threshold: float,
    ):
        """
        Mean Average Precision (mAP) metric for object detection tasks.

        The AP computation uses 101-point interpolation, which calculates the average
        precision for each class by interpolating the precision-recall curve at 101 evenly
        spaced recall levels from 0 to 1. The mAP is then calculated by averaging these
        values across all class labels.

        Parameters
        ----------
        value : float
            The mean average precision value.
        iou_threshold : float
            The IOU threshold used to compute the mAP.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.mAP.value,
            value=value,
            parameters={
                "iou_threshold": iou_threshold,
            },
        )

    @classmethod
    def average_precision_averaged_over_IOUs(
        cls,
        value: float,
        iou_thresholds: list[float],
        label: str,
    ):
        """
        Average Precision (AP) metric averaged over multiple IOU thresholds.

        The AP computation uses 101-point interpolation, which calculates the average precision
        by interpolating the precision-recall curve at 101 evenly spaced recall levels from 0 to 1
        for each IOU threshold specified in `iou_thresholds`. The final APAveragedOverIOUs value is
        obtained by averaging these AP values across all specified IOU thresholds.

        Parameters
        ----------
        value : float
            The average precision value averaged over the specified IOU thresholds.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the AP values.
        label : str
            The class label for which the AP is computed.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.APAveragedOverIOUs.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "label": label,
            },
        )

    @classmethod
    def mean_average_precision_averaged_over_IOUs(
        cls,
        value: float,
        iou_thresholds: list[float],
    ):
        """
        Mean Average Precision (mAP) metric averaged over multiple IOU thresholds.

        The AP computation uses 101-point interpolation, which calculates the average precision
        by interpolating the precision-recall curve at 101 evenly spaced recall levels from 0 to 1
        for each IOU threshold specified in `iou_thresholds`. The final mAPAveragedOverIOUs value is
        obtained by averaging these AP values across all specified IOU thresholds and all class labels.

        Parameters
        ----------
        value : float
            The average precision value averaged over the specified IOU thresholds.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the AP values.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.mAPAveragedOverIOUs.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
            },
        )

    @classmethod
    def average_recall(
        cls,
        value: float,
        score_threshold: float,
        iou_thresholds: list[float],
        label: str,
    ):
        """
        Average Recall (AR) metric for object detection tasks.

        The AR computation considers detections with confidence scores above the specified
        `score_threshold` and calculates the recall at each IOU threshold in `iou_thresholds`.
        The final AR value is the average of these recall values across all specified IOU
        thresholds.

        Parameters
        ----------
        value : float
            The average recall value averaged over the specified IOU thresholds.
        score_threshold : float
            The detection score threshold; only detections with confidence scores above this
            threshold are considered.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the recall values.
        label : str
            The class label for which the AR is computed.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.AR.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "score_threshold": score_threshold,
                "label": label,
            },
        )

    @classmethod
    def mean_average_recall(
        cls,
        value: float,
        score_threshold: float,
        iou_thresholds: list[float],
    ):
        """
        Mean Average Recall (mAR) metric for object detection tasks.

        The mAR computation considers detections with confidence scores above the specified
        `score_threshold` and calculates recall at each IOU threshold in `iou_thresholds` for
        each label. The final mAR value is obtained by averaging these recall values over the
        specified IOU thresholds and then averaging across all labels.

        Parameters
        ----------
        value : float
            The mean average recall value averaged over the specified IOU thresholds.
        score_threshold : float
            The detection score threshold; only detections with confidence scores above this
            threshold are considered.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the recall values.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.mAR.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "score_threshold": score_threshold,
            },
        )

    @classmethod
    def average_recall_averaged_over_scores(
        cls,
        value: float,
        score_thresholds: list[float],
        iou_thresholds: list[float],
        label: str,
    ):
        """
        Average Recall (AR) metric averaged over multiple score thresholds for a specific object class label.

        The AR computation considers detections across multiple `score_thresholds` and calculates
        recall at each IOU threshold in `iou_thresholds`. The final AR value is obtained by averaging
        the recall values over all specified score thresholds and IOU thresholds.

        Parameters
        ----------
        value : float
            The average recall value averaged over the specified score thresholds and IOU thresholds.
        score_thresholds : list[float]
            The list of detection score thresholds; detections with confidence scores above each threshold are considered.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the recall values.
        label : str
            The class label for which the AR is computed.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.ARAveragedOverScores.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "score_thresholds": score_thresholds,
                "label": label,
            },
        )

    @classmethod
    def mean_average_recall_averaged_over_scores(
        cls,
        value: float,
        score_thresholds: list[float],
        iou_thresholds: list[float],
    ):
        """
        Mean Average Recall (mAR) metric averaged over multiple score thresholds and IOU thresholds.

        The mAR computation considers detections across multiple `score_thresholds`, calculates recall
        at each IOU threshold in `iou_thresholds` for each label, averages these recall values over all
        specified score thresholds and IOU thresholds, and then computes the mean across all labels to
        obtain the final mAR value.

        Parameters
        ----------
        value : float
            The mean average recall value averaged over the specified score thresholds and IOU thresholds.
        score_thresholds : list[float]
            The list of detection score thresholds; detections with confidence scores above each threshold are considered.
        iou_thresholds : list[float]
            The list of IOU thresholds used to compute the recall values.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.mARAveragedOverScores.value,
            value=value,
            parameters={
                "iou_thresholds": iou_thresholds,
                "score_thresholds": score_thresholds,
            },
        )

    @classmethod
    def precision_recall_curve(
        cls,
        precisions: list[float],
        scores: list[float],
        iou_threshold: float,
        label: str,
    ):
        """
        Interpolated precision-recall curve over 101 recall points.

        The precision values are interpolated over recalls ranging from 0.0 to 1.0 in steps of 0.01,
        resulting in 101 points. This is a byproduct of the 101-point interpolation used in calculating
        the Average Precision (AP) metric in object detection tasks.

        Parameters
        ----------
        precisions : list[float]
            Interpolated precision values corresponding to recalls at 0.0, 0.01, ..., 1.0.
        scores : list[float]
            Maximum prediction score for each point on the interpolated curve.
        iou_threshold : float
            The Intersection over Union (IOU) threshold used to determine true positives.
        label : str
            The class label associated with this precision-recall curve.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.PrecisionRecallCurve.value,
            value={
                "precisions": precisions,
                "scores": scores,
            },
            parameters={
                "iou_threshold": iou_threshold,
                "label": label,
            },
        )

    @classmethod
    def counts(
        cls,
        tp: int,
        fp: int,
        fn: int,
        label: str,
        iou_threshold: float,
        score_threshold: float,
    ):
        """
        `Counts` encapsulates the counts of true positives (`tp`), false positives (`fp`),
        and false negatives (`fn`) for object detection evaluation, along with the associated
        class label, Intersection over Union (IOU) threshold, and confidence score threshold.

        Parameters
        ----------
        tp : int
            Number of true positives.
        fp : int
            Number of false positives.
        fn : int
            Number of false negatives.
        label : str
            The class label for which the counts are calculated.
        iou_threshold : float
            The IOU threshold used to determine a match between predicted and ground truth boxes.
        score_threshold : float
            The confidence score threshold above which predictions are considered.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Counts.value,
            value={
                "tp": tp,
                "fp": fp,
                "fn": fn,
            },
            parameters={
                "iou_threshold": iou_threshold,
                "score_threshold": score_threshold,
                "label": label,
            },
        )

    @classmethod
    def confusion_matrix(
        cls,
        confusion_matrix: dict[
            str,  # ground truth label value
            dict[
                str,  # prediction label value
                dict[
                    str,  # either `count` or `examples`
                    int
                    | list[
                        dict[
                            str,  # either `datum`, `groundtruth`, `prediction` or score
                            str  # datum uid
                            | dict[
                                str, float
                            ]  # bounding box (xmin, xmax, ymin, ymax)
                            | float,  # prediction score
                        ]
                    ],
                ],
            ],
        ],
        hallucinations: dict[
            str,  # prediction label value
            dict[
                str,  # either `count` or `examples`
                int
                | list[
                    dict[
                        str,  # either `datum`, `prediction` or score
                        str  # datum uid
                        | float  # prediction score
                        | dict[
                            str, float
                        ],  # bounding box (xmin, xmax, ymin, ymax)
                    ]
                ],
            ],
        ],
        missing_predictions: dict[
            str,  # ground truth label value
            dict[
                str,  # either `count` or `examples`
                int
                | list[
                    dict[
                        str,  # either `datum` or `groundtruth`
                        str  # datum uid
                        | dict[
                            str, float
                        ],  # bounding box (xmin, xmax, ymin, ymax)
                    ]
                ],
            ],
        ],
        score_threshold: float,
        iou_threshold: float,
        maximum_number_of_examples: int,
    ):
        """
        Confusion matrix for object detection tasks.

        This class encapsulates detailed information about the model's performance, including correct
        predictions, misclassifications, hallucinations (false positives), and missing predictions
        (false negatives). It provides counts and examples for each category to facilitate in-depth analysis.

        Confusion Matrix Format:
        {
            <ground truth label>: {
                <prediction label>: {
                    'count': int,
                    'examples': [
                        {
                            'datum': str,
                            'groundtruth': dict,  # {'xmin': float, 'xmax': float, 'ymin': float, 'ymax': float}
                            'prediction': dict,   # {'xmin': float, 'xmax': float, 'ymin': float, 'ymax': float}
                            'score': float,
                        },
                        ...
                    ],
                },
                ...
            },
            ...
        }

        Hallucinations Format:
        {
            <prediction label>: {
                'count': int,
                'examples': [
                    {
                        'datum': str,
                        'prediction': dict,  # {'xmin': float, 'xmax': float, 'ymin': float, 'ymax': float}
                        'score': float,
                    },
                    ...
                ],
            },
            ...
        }

        Missing Prediction Format:
        {
            <ground truth label>: {
                'count': int,
                'examples': [
                    {
                        'datum': str,
                        'groundtruth': dict,  # {'xmin': float, 'xmax': float, 'ymin': float, 'ymax': float}
                    },
                    ...
                ],
            },
            ...
        }

        Parameters
        ----------
        confusion_matrix : dict
            A nested dictionary where the first key is the ground truth label value, the second key
            is the prediction label value, and the innermost dictionary contains either a `count`
            or a list of `examples`. Each example includes the datum UID, ground truth bounding box,
            predicted bounding box, and prediction scores.
        hallucinations : dict
            A dictionary where each key is a prediction label value with no corresponding ground truth
            (false positives). The value is a dictionary containing either a `count` or a list of
            `examples`. Each example includes the datum UID, predicted bounding box, and prediction score.
        missing_predictions : dict
            A dictionary where each key is a ground truth label value for which the model failed to predict
            (false negatives). The value is a dictionary containing either a `count` or a list of `examples`.
            Each example includes the datum UID and ground truth bounding box.
        score_threshold : float
            The confidence score threshold used to filter predictions.
        iou_threshold : float
            The Intersection over Union (IOU) threshold used to determine true positives.
        maximum_number_of_examples : int
            The maximum number of examples per element.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.ConfusionMatrix.value,
            value={
                "confusion_matrix": confusion_matrix,
                "hallucinations": hallucinations,
                "missing_predictions": missing_predictions,
            },
            parameters={
                "score_threshold": score_threshold,
                "iou_threshold": iou_threshold,
                "maximum_number_of_examples": maximum_number_of_examples,
            },
        )



import json

import pytest
from valor_lite.text_generation import Context, QueryResponse
from valor_lite.text_generation.llm.integrations import _validate_messages


class MockWrapper:
    """
    A mocked LLM client for testing purposes.
    """

    def __init__(
        self,
        **kwargs,
    ):
        self.model_name = "mock"
        self.returning = "test payload"

    def _process_messages(
        self,
        messages: list[dict[str, str]],
    ) -> list[dict[str, str]]:
        _validate_messages(messages=messages)  # type: ignore
        return messages

    def __call__(
        self,
        messages: list[dict[str, str]],
    ) -> str:
        if isinstance(self.returning, str):
            return self.returning
        elif isinstance(self.returning, list) and len(self.returning) > 0:
            return self.returning.pop()
        return "null"


@pytest.fixture
def mock_client():
    return MockWrapper()


@pytest.fixture
def rag_queries() -> list[str]:
    return [
        """Did John Adams get along with Alexander Hamilton?""",
        """Did Lincoln win the election of 1860?""",
        """If a turtle egg was kept warm, what would likely hatch?""",
    ]


@pytest.fixture
def rag_predictions() -> list[str]:
    return [
        """Based on the provided context, John Adams and Alexander Hamilton did not get along. John Adams, during his presidency, had grown independent of his cabinet, often making decisions despite opposition from it. Hamilton, who was accustomed to being regularly consulted by Washington, sent Adams a detailed letter with policy suggestions after his inauguration, which Adams dismissively ignored.\n""",
        """Yes, Lincoln won the election of 1860. He received the highest number of votes and a majority in the Electoral College, making him the 16th President of the United States. However, it's important to note that he won entirely due to his support in the North and West, as he did not receive any votes in 10 of the 15 Southern slave states.""",
        """If a turtle egg was kept warm, it would likely hatch into a baby turtle. The sex of the baby turtle would be determined by the incubation temperature, assuming the species is one of those that determine sex thermally. This is because many turtle species have the ability to move around inside their eggs to select the best temperature for development, which can influence their sexual destiny.""",
    ]


@pytest.fixture
def rag_references() -> list[str]:
    return [
        """John Adams and Alexander Hamilton did not get along. John Adams had grown independent of his cabinet, often making decisions despite opposition from it.\n""",  # same as prediction with some strings deleted
        """Yes, Lincoln won the election of 1860. He received the highest number of votes and a majority in the Electoral College, making him the 16th President of the United States. However, it's important to note that he won entirely due to his support in the North and West, as he did not receive any votes in 10 of the 15 Southern slave states.""",  # same as prediction
        """If kept warm, it would hatch a coyote.""",  # very different than prediction
    ]


@pytest.fixture
def rag_context() -> list[list[str]]:
    return [
        [
            """Although aware of Hamilton\'s influence, Adams was convinced that their retention ensured a smoother succession. Adams maintained the economic programs of Hamilton, who regularly consulted with key cabinet members, especially the powerful Treasury Secretary, Oliver Wolcott Jr. Adams was in other respects quite independent of his cabinet, often making decisions despite opposition from it. Hamilton had grown accustomed to being regularly consulted by Washington. Shortly after Adams was inaugurated, Hamilton sent him a detailed letter with policy suggestions. Adams dismissively ignored it.\n\nFailed peace commission and XYZ affair\nHistorian Joseph Ellis writes that "[t]he Adams presidency was destined to be dominated by a single question of American policy to an extent seldom if ever encountered by any succeeding occupant of the office." That question was whether to make war with France or find peace. Britain and France were at war as a result of the French Revolution. Hamilton and the Federalists strongly favored the British monarchy against what they denounced as the political radicalism and anti-religious frenzy of the French Revolution. Jefferson and the Republicans, with their firm opposition to monarchy, strongly supported the French overthrowing their king. The French had supported Jefferson for president in 1796 and became belligerent at his loss.""",
            """Led by Revolutionary War veteran John Fries, rural German-speaking farmers protested what they saw as a threat to their liberties. They intimidated tax collectors, who often found themselves unable to go about their business. The disturbance was quickly ended with Hamilton leading the army to restore peace.Fries and two other leaders were arrested, found guilty of treason, and sentenced to hang. They appealed to Adams requesting a pardon. The cabinet unanimously advised Adams to refuse, but he instead granted the pardon, arguing the men had instigated a mere riot as opposed to a rebellion. In his pamphlet attacking Adams before the election, Hamilton wrote that \"it was impossible to commit a greater error.\"\n\nFederalist divisions and peace\nOn May 5, 1800, Adams's frustrations with the Hamilton wing of the party exploded during a meeting with McHenry, a Hamilton loyalist who was universally regarded, even by Hamilton, as an inept Secretary of War. Adams accused him of subservience to Hamilton and declared that he would rather serve as Jefferson's vice president or minister at The Hague than be beholden to Hamilton for the presidency. McHenry offered to resign at once, and Adams accepted. On May 10, he asked Pickering to resign.""",
            """Indeed, Adams did not consider himself a strong member of the Federalist Party. He had remarked that Hamilton\'s economic program, centered around banks, would "swindle" the poor and unleash the "gangrene of avarice." Desiring "a more pliant president than Adams," Hamilton maneuvered to tip the election to Pinckney. He coerced South Carolina Federalist electors, pledged to vote for "favorite son" Pinckney, to scatter their second votes among candidates other than Adams. Hamilton\'s scheme was undone when several New England state electors heard of it and agreed not to vote for Pinckney. Adams wrote shortly after the election that Hamilton was a "proud Spirited, conceited, aspiring Mortal always pretending to Morality, with as debauched Morals as old Franklin who is more his Model than any one I know." Throughout his life, Adams made highly critical statements about Hamilton. He made derogatory references to his womanizing, real or alleged, and slurred him as the "Creole bastard.""",
            """The pair\'s exchange was respectful; Adams promised to do all that he could to restore friendship and cordiality "between People who, tho Seperated [sic] by an Ocean and under different Governments have the Same Language, a Similar Religion and kindred Blood," and the King agreed to "receive with Pleasure, the Assurances of the friendly Dispositions of the United States." The King added that although "he had been the last to consent" to American independence, he had always done what he thought was right. He startled Adams by commenting that "There is an Opinion, among Some People, that you are not the most attached of all Your Countrymen, to the manners of France." Adams replied, "That Opinion sir, is not mistaken... I have no Attachments but to my own Country." King George responded, "An honest Man will never have any other."\nAdams was joined by Abigail in London. Suffering the hostility of the King\'s courtiers, they escaped when they could by seeking out Richard Price, minister of Newington Green Unitarian Church and instigator of the debate over the Revolution within Britain.""",
        ],
        [
            """Republican speakers focused first on the party platform, and second on Lincoln's life story, emphasizing his childhood poverty. The goal was to demonstrate the power of \"free labor\", which allowed a common farm boy to work his way to the top by his own efforts. The Republican Party's production of campaign literature dwarfed the combined opposition; a Chicago Tribune writer produced a pamphlet that detailed Lincoln's life and sold 100,000\u2013200,000 copies. Though he did not give public appearances, many sought to visit him and write him. In the runup to the election, he took an office in the Illinois state capitol to deal with the influx of attention. He also hired John George Nicolay as his personal secretary, who would remain in that role during the presidency.On November 6, 1860, Lincoln was elected the 16th president. He was the first Republican president and his victory was entirely due to his support in the North and West. No ballots were cast for him in 10 of the 15 Southern slave states, and he won only two of 996 counties in all the Southern states, an omen of the impending Civil War.""",
            """Lincoln received 1,866,452 votes, or 39.8% of the total in a four-way race, carrying the free Northern states, as well as California and Oregon. His victory in the Electoral College was decisive: Lincoln had 180 votes to 123 for his opponents.\n\nPresidency (1861\u20131865)\nSecession and inauguration\nThe South was outraged by Lincoln's election, and in response secessionists implemented plans to leave the Union before he took office in March 1861. On December 20, 1860, South Carolina took the lead by adopting an ordinance of secession; by February 1, 1861, Florida, Mississippi, Alabama, Georgia, Louisiana, and Texas followed. Six of these states declared themselves to be a sovereign nation, the Confederate States of America, and adopted a constitution. The upper South and border states (Delaware, Maryland, Virginia, North Carolina, Tennessee, Kentucky, Missouri, and Arkansas) initially rejected the secessionist appeal. President Buchanan and President-elect Lincoln refused to recognize the Confederacy, declaring secession illegal.""",
            """In 1860, Lincoln described himself: "I am in height, six feet, four inches, nearly; lean in flesh, weighing, on an average, one hundred and eighty pounds; dark complexion, with coarse black hair, and gray eyes." Michael Martinez wrote about the effective imaging of Lincoln by his campaign. At times he was presented as the plain-talking "Rail Splitter" and at other times he was "Honest Abe", unpolished but trustworthy.On May 18, at the Republican National Convention in Chicago, Lincoln won the nomination on the third ballot, beating candidates such as Seward and Chase. A former Democrat, Hannibal Hamlin of Maine, was nominated for vice president to balance the ticket. Lincoln\'s success depended on his campaign team, his reputation as a moderate on the slavery issue, and his strong support for internal improvements and the tariff. Pennsylvania put him over the top, led by the state\'s iron interests who were reassured by his tariff support. Lincoln\'s managers had focused on this delegation while honoring Lincoln\'s dictate to "Make no contracts that will bind me".As the Slave Power tightened its grip on the national government, most Republicans agreed with Lincoln that the North was the aggrieved party.""",
            """The Confederate government evacuated Richmond and Lincoln visited the conquered capital. On April 9, Lee surrendered to Grant at Appomattox, officially ending the war.\n\nReelection\nLincoln ran for reelection in 1864, while uniting the main Republican factions, along with War Democrats Edwin M. Stanton and Andrew Johnson. Lincoln used conversation and his patronage powers\u2014greatly expanded from peacetime\u2014to build support and fend off the Radicals' efforts to replace him. At its convention, the Republicans selected Johnson as his running mate. To broaden his coalition to include War Democrats as well as Republicans, Lincoln ran under the label of the new Union Party.\nGrant's bloody stalemates damaged Lincoln's re-election prospects, and many Republicans feared defeat. Lincoln confidentially pledged in writing that if he should lose the election, he would still defeat the Confederacy before turning over the White House; Lincoln did not show the pledge to his cabinet, but asked them to sign the sealed envelope. The pledge read as follows:This morning, as for some days past, it seems exceedingly probable that this Administration will not be re-elected.""",
        ],
        [
            """There is experimental evidence that the embryos of Mauremys reevesii can move around inside their eggs to select the best temperature for development, thus influencing their sexual destiny. In other species, sex is determined genetically. The length of incubation for turtle eggs varies from two to three months for temperate species, and four months to over a year for tropical species. Species that live in warm temperate climates can delay their development.Hatching young turtles break out of the shell using an egg tooth, a sharp projection that exists temporarily on their upper beak. Hatchlings dig themselves out of the nest and find safety in vegetation or water. Some species stay in the nest for longer, be it for overwintering or to wait for the rain to loosen the soil for them to dig out. Young turtles are highly vulnerable to predators, both in the egg and as hatchlings. Mortality is high during this period but significantly decreases when they reach adulthood. Most species grow quickly during their early years and slow down when they are mature.\n\nLifespan\nTurtles can live long lives.""",
            """Females usually dig a flask-like chamber in the substrate. Other species lay their eggs in vegetation or crevices. Females choose nesting locations based on environmental factors such as temperature and humidity, which are important for developing embryos. Depending on the species, the number of eggs laid varies from one to over 100. Larger females can lay eggs that are greater in number or bigger in size. Compared to freshwater turtles, tortoises deposit fewer but larger eggs. Females can lay multiple clutches throughout a season, particularly in species that experience unpredictable monsoons.\nMost mother turtles do no more in the way of parental care than covering their eggs and immediately leaving, though some species guard their nests for days or weeks. Eggs vary between rounded, oval, elongated, and between hard- and soft-shelled. Most species have their sex determined by temperature. In some species, higher temperatures produce females and lower ones produce males, while in others, milder temperatures produce males and both hot and cold extremes produce females.""",
            """In species like the Russian tortoise, the male has a lighter shell and longer legs. The high, rounded shape of box turtles are particular obstacles for mounting. The male eastern box turtle leans backward and hooks onto the back of the female's plastron. Aquatic turtles mount in water, and female sea turtles support the mounting male while swimming and diving. During copulation, the male turtle aligns his tail with the female's so he can insert his penis into her cloaca. Some female turtles can store sperm from multiple males and their egg clutches can have multiple sires.\n\nEggs and hatchlings\nTurtles, including sea turtles, lay their eggs on land, although some lay eggs near water that rises and falls in level, submerging the eggs. While most species build nests and lay eggs where they forage, some travel miles. The common snapping turtle walks 5 km (3 mi) on land, while sea turtles travel even further; the leatherback swims some 12,000 km (7,500 mi) to its nesting beaches. Most turtles create a nest for their eggs. Females usually dig a flask-like chamber in the substrate.""",
            """Turtles are ectotherms or \"cold-blooded\", meaning that their internal temperature varies with their direct environment. They are generally opportunistic omnivores and mainly feed on plants and animals with limited movements. Many turtles migrate short distances seasonally. Sea turtles are the only reptiles that migrate long distances to lay their eggs on a favored beach.\nTurtles have appeared in myths and folktales around the world. Some terrestrial and freshwater species are widely kept as pets. Turtles have been hunted for their meat, for use in traditional medicine, and for their shells. Sea turtles are often killed accidentally as bycatch in fishing nets. Turtle habitats around the world are being destroyed. As a result of these pressures, many species are extinct or threatened with extinction.\n\nNaming and etymology\nThe word turtle is borrowed from the French word tortue or tortre 'turtle, tortoise'. It is a common name and may be used without knowledge of taxonomic distinctions. In North America, it may denote the order as a whole. In Britain, the name is used for sea turtles as opposed to freshwater terrapins and land-dwelling tortoises.""",
        ],
    ]


@pytest.fixture
def content_gen_queries() -> list[str]:
    return [
        """Write about a haunted house from the perspective of the ghost.""",
        """Explain to an elementary school student how to do long multiplication with the example 43 times 22. The resulting answer should be 946.""",
        """Draft an email to a coworker explaining a project delay. Explain that the delay is due to funding cuts, which resulted in multiple employees being moved to different projects. Inform the coworker that the project deadline will have to be pushed back. Be apologetic and professional. Express eagerness to still complete the project as efficiently as possible.""",
    ]


@pytest.fixture
def content_gen_predictions() -> list[str]:
    return [
        """As a ghost haunting the old, decrepit house on Elm Street, I am trapped in a state of perpetual torment and longing. I drift through the dusty halls, my translucent figure flickering in and out of existence as I relive the memories of my past life.\n\nMy presence is felt by those who dare to enter the house, their hairs standing on end as they sense the chill in the air and the whispers that echo through the rooms. I watch as fear grips their hearts, knowing that I am the reason for their unease.\n\nI am bound to this house by unfinished business, a deep-rooted need for closure that eludes me even in death. I long to reach out to the living, to make them understand the pain and sorrow that consume me, but my ethereal form cannot touch them.\n\nYet, despite the fear and dread that my presence evokes, there is a part of me that yearns for connection, for someone to see beyond the horror and recognize the lost soul that I am. But until that day comes, I remain a ghost trapped within the walls of the haunted house on Elm Street, my existence a haunting reminder of the past that refuses to fade away.""",
        """Sure! To multiply 43 by 22, you need to follow these steps:\n\n1. Write down the numbers vertically, with one on top of the other:\n\t\n```\t\n   43\t\nx 22\t\n-----\t\n```\n\n2. Start by multiplying the rightmost digit of the bottom number (2) by each digit in the top number. Write down each result below the line, leaving a space for the next step:\n\t\n```\t\n   43\t\nx 22\t\n-----\t\n   86\t\n```\n\n3. Next, move to the left and multiply the left digit of the bottom number (2) by each digit in the top number. Write down each result with one position shifted to the left from the previous step. Add any results from previous steps that overlap:\n\t\n```\t\n  43\t\nx 22\t\n-----\t\n  86\t\n+86\t\n-----\t\n946\t\n```\n\nSo, 43 times 22 is equal to 946.""",
        """Subject: Project Delay Due to Funding Cuts\n\nDear [Coworker's Name],\n\nI hope this message finds you well. I am writing to update you on the status of our project and unfortunately, convey some disappointing news.\n\nDue to recent funding cuts within our department, we have had to make some adjustments to project assignments. As a result, multiple employees, including key team members for our current project, have been moved to different projects to accommodate the changes. This unexpected shift has impacted our project timeline.\n\nI regret to inform you that our project deadline will need to be pushed back in light of these developments. I understand the inconvenience this may cause and I sincerely apologize for any disruption this may cause to your schedule or other commitments.\n\nPlease rest assured that despite these unforeseen circumstances, I am fully committed to completing the project efficiently and effectively. I will work closely with the team to develop a revised timeline and ensure that we deliver quality work that meets our objectives.\n\nThank you for your understanding and continued support during this challenging period. I value your collaboration and look forward to working together to overcome this setback and achieve our project goals.\n\nIf you have any questions or concerns, please feel free to reach out to me. I appreciate your patience as we navigate through this situation together.\n\nBest regards,\n\n[Your Name]""",
    ]


@pytest.fixture
def summarization_texts() -> list[str]:
    return [
        """Aston Villa take on Liverpool in their FA Cup semi-final encounter on Sunday with the competition both sides' last chance to win any silverware this season. Sportsmail columnist Jamie Redknapp looks ahead to the Wembley showdown and where the match could be won and lost with individual player duels. CHRISTIAN BENTEKE v MARTIN SKRTEL . This will be a heavyweight contest that could decide the game. Christian Benteke is superb in the air and Martin Skrtel will have his hands full. Liverpool have to stop the supply line because defending crosses has been their Achilles heel this season. Christian Benteke (centre) scored the only goal of the game as Villa won 1-0 at Tottenham on April 11 . Liverpool defender Martin Skrtel (right) will have his hands full trying to stop Benteke on Sunday afternoon . FABIAN DELPH v JORDAN HENDERSON . This should be a good contest between two England team-mates. Fabian Delph’s new deal was a real boost for Villa - he drives that midfield, though he doesn’t get enough goals. You used to say the same about Jordan Henderson but he has improved so much. England international Fabian Delph (left) and Jordan Henderson are set for a midfield battle at Wembley . RAHEEM STERLING v RON VLAAR and NATHAN BAKER . Ron Vlaar and Nathan Baker make an imposing back line but they would rather be up against a Benteke than a Raheem Sterling, who will float around and make himself difficult to mark so he can use his lightning pace to get in behind them. Raheem Sterling's (left) pace and trickery is bound to cause the Villa defence a lot of problems . Ron Vlaar (left) was part of the Villa defence that kept a clean sheet at Spurs in the Premier League . The Holland international and Nathan Baker (right) will be hoping to do likewise against the Reds at Wembley.""",
        """Juventus and Liverpool are continuing to monitor developments with Chelsea midfielder Oscar. The Brazil international has been criticised by Jose Mourinho in recent weeks and there are question marks over his future. Chelsea want to strengthen in the summer and may need a high profile departure to help balance the books. Juventus and Liverpool are interested in signing Chelsea 23-year-old midfielder Oscar . Oscar in action during Chelsea's 1-0 Premier League victory against Queens Park Rangers last weekend . Oscar cost Chelsea £19.35m and they would want a substantial profit on the 23 year-old. Paris Saintt Germain have shown interest in the past also. Juventus want a playmaker for next season and Brazil boss Carlos Dunga advised them to buy Oscar. 'He reminds me of Roberto Baggio,' he said. 'Oscar has technique, reads situations well and is a modern and versatile trequartista. He reminds me of Roberto Baggio, but also has similarities to Massimiliano Allegri. The former Sao Paulo youngster has struggled to make an impact for Chelsea this season . Brazil coach Dunga (pictured) revealed the Chelsea midfielder reminds him of Roberto Baggio . 'Brazilians like to have fun with their football, which hasn’t happened to Oscar very much recently, but I met Jose Mourinho and he spoke highly of all his Brazilian players. 'I tell Allegri that Oscar is strong and also a good lad. A forward line with him, Carlos Tevez and Alvaro Morata would drive any Coach crazy. 'It wouldn’t be a step backwards for Oscar to go to Juventus. He’d be decisive in Serie A and whether he plays for Juventus or Chelsea it’ll always be a great club.' Oscar celebrates scoring Chelsea's fourth goal during the 5-0 victory against Swansea in January.""",
    ]


@pytest.fixture
def summarization_predictions() -> list[str]:
    return [
        """Aston Villa and Liverpool face off in the FA Cup semi-final as both teams look to secure their last chance at silverware this season. Sportsmail columnist Jamie Redknapp analyzes key player duels that could decide the game, such as Christian Benteke against Martin Skrtel, Fabian Delph against Jordan Henderson, and Raheem Sterling against Ron Vlaar and Nathan Baker. Redknapp emphasizes the importance of stopping the supply line to Benteke and dealing with Sterling's pace and trickery in the match.""",
        """Juventus and Liverpool are showing interest in Chelsea midfielder Oscar, who has faced criticism and uncertainty about his future at the club. Chelsea may need to sell a high-profile player to strengthen their squad in the summer. Oscar, who was signed for £19.35m, has also attracted interest from Paris Saint-Germain in the past. Brazil coach Carlos Dunga sees qualities in Oscar similar to Roberto Baggio and believes he could be a key player for Juventus.""",
    ]


@pytest.fixture
def rag_annotations(
    rag_queries: list[str],
    rag_context: list[list[str]],
    rag_predictions: list[str],
    rag_references: list[str],
) -> list[QueryResponse]:
    return [
        QueryResponse(
            query=rag_queries[i],
            response=rag_predictions[i],
            context=Context(
                groundtruth=[
                    rag_references[i],
                    "some other text",
                    "some final text",
                ],
                prediction=rag_context[i],
            ),
        )
        for i in range(len(rag_queries))
    ]


@pytest.fixture
def content_gen_annotations(
    content_gen_queries: list[str], content_gen_predictions: list[str], content
) -> list[QueryResponse]:
    return [
        QueryResponse(
            query=content_gen_queries[i],
            response=content_gen_predictions[i],
        )
        for i in range(len(content_gen_queries))
    ]


@pytest.fixture
def summarization_annotations(
    summarization_texts: list[str],
    summarization_predictions: list[str],
) -> list[QueryResponse]:
    return [
        QueryResponse(
            query=summarization_texts[i],
            response=summarization_predictions[i],
        )
        for i in range(len(summarization_texts))
    ]


@pytest.fixture
def verdicts_all_yes() -> str:
    return json.dumps(
        {
            "verdicts": [
                {"verdict": "yes", "analysis": "some text"},
                {"verdict": "yes", "analysis": "some text"},
                {"verdict": "yes", "analysis": "some text"},
            ],
            "statements": ["x", "y", "z"],
            "opinions": ["x", "y", "z"],
            "claims": ["x", "y", "z"],
        }
    )


@pytest.fixture
def verdicts_all_no() -> str:
    return json.dumps(
        {
            "verdicts": [
                {"verdict": "no", "analysis": "some text"},
                {"verdict": "no", "analysis": "some text"},
                {"verdict": "no", "analysis": "some text"},
            ],
            "statements": ["x", "y", "z"],
            "opinions": ["x", "y", "z"],
            "claims": ["x", "y", "z"],
        }
    )


@pytest.fixture
def verdicts_two_yes_one_no() -> str:
    return json.dumps(
        {
            "verdicts": [
                {"verdict": "yes", "analysis": "some text"},
                {"verdict": "no", "analysis": "some text"},
                {"verdict": "yes", "analysis": "some text"},
            ],
            "statements": ["x", "y", "z"],
            "opinions": ["x", "y", "z"],
            "claims": ["x", "y", "z"],
            "TP": ["x", "y"],
            "FP": ["z"],
            "FN": [],
            "unused": 4,
        }
    )


@pytest.fixture
def verdicts_empty() -> str:
    return json.dumps(
        {
            "verdicts": [],
            "statements": [],
            "opinions": [],
            "claims": [],
        }
    )



import numpy as np
import pytest
from shapely.geometry import Polygon as ShapelyPolygon
from valor_lite.object_detection import (
    Bitmask,
    BoundingBox,
    Detection,
    Polygon,
)


def test_BoundingBox():
    # groundtruth
    gt = BoundingBox(xmin=0, xmax=1, ymin=0, ymax=1, labels=["label"])

    # prediction
    pd = BoundingBox(
        xmin=-1,
        xmax=11,
        ymin=0,
        ymax=1,
        labels=["label"],
        scores=[0.7],
    )

    with pytest.raises(ValueError):
        BoundingBox(
            xmin=0,
            xmax=1,
            ymin=0,
            ymax=1,
            labels=["label"],
            scores=[0.7, 0.1],
        )
    with pytest.raises(ValueError):
        BoundingBox(
            xmin=0,
            xmax=1,
            ymin=0,
            ymax=1,
            labels=["label1", "label2"],
            scores=[0.7],
        )

    # test `extrema` property
    assert gt.extrema == (0, 1, 0, 1)
    assert pd.extrema == (-1, 11, 0, 1)


def test_Bitmask():

    mask = mask = np.zeros((10, 10), dtype=np.bool_)
    mask[:5, :5] = True

    # groundtruth
    gt = Bitmask(mask=mask, labels=["label"])

    # prediction
    Bitmask(
        mask=mask,
        labels=["label"],
        scores=[0.7],
    )

    # test score-label matching
    with pytest.raises(ValueError):
        Bitmask(
            mask=np.zeros((10, 10), dtype=np.bool_),
            labels=["label"],
            scores=[0.7, 0.1],
        )
    with pytest.raises(ValueError):
        Bitmask(
            mask=np.zeros((10, 10), dtype=np.bool_),
            labels=["label1", "label2"],
            scores=[0.7],
        )

    # test `to_box` method
    assert gt.extrema == (0, 4, 0, 4)

    with pytest.raises(ValueError):
        Bitmask(mask=np.array([], dtype=np.bool_), labels=["label"])


def test_Polygon(rect1_rotated_5_degrees_around_origin):

    shape = ShapelyPolygon(rect1_rotated_5_degrees_around_origin)

    # groundtruth
    gt = Polygon(shape=shape, labels=["label"])

    # prediction
    Polygon(
        shape=shape,
        labels=["label"],
        scores=[0.7],
    )

    # test score-label matching
    with pytest.raises(ValueError):
        Polygon(
            shape=shape,
            labels=["label"],
            scores=[0.7, 0.1],
        )
    with pytest.raises(ValueError):
        Polygon(
            shape=shape,
            labels=["label1", "label2"],
            scores=[0.7],
        )
    # test that we throw a type error if the shape isn't a shapely.geometry.Polygon
    with pytest.raises(TypeError):
        Polygon(
            shape=np.zeros((10, 10), dtype=np.bool_),  # type: ignore - purposefully throwing error
            labels=["label1", "label2"],
            scores=[0.7],
        )

    # test `to_box` method
    assert gt.extrema == (
        6.475717271011129,
        58.90012445802815,
        10.833504408394036,
        45.07713248852931,
    )

    with pytest.raises(ValueError):
        Polygon(shape=ShapelyPolygon([]), labels=["label"])


def test_Detection():

    # groundtruth
    gt = BoundingBox(xmin=0, xmax=1, ymin=0, ymax=1, labels=["label"])

    # prediction
    pd = BoundingBox(
        xmin=-1,
        xmax=11,
        ymin=0,
        ymax=1,
        labels=["label"],
        scores=[0.7],
    )

    Detection(
        uid="uid",
        groundtruths=[gt],
        predictions=[pd],
    )

    # test that predictions must contain scores
    with pytest.raises(ValueError):
        Detection(
            uid="uid",
            groundtruths=[gt],
            predictions=[gt],
        )



from dataclasses import dataclass
from enum import Enum

from valor_lite.schemas import BaseMetric


class MetricType(Enum):
    Precision = "Precision"
    Recall = "Recall"
    Accuracy = "Accuracy"
    F1 = "F1"
    IOU = "IOU"
    mIOU = "mIOU"
    ConfusionMatrix = "ConfusionMatrix"


@dataclass
class Metric(BaseMetric):
    """
    Semantic Segmentation Metric.

    Attributes
    ----------
    type : str
        The metric type.
    value : int | float | dict
        The metric value.
    parameters : dict[str, Any]
        A dictionary containing metric parameters.
    """

    def __post_init__(self):
        if not isinstance(self.type, str):
            raise TypeError(
                f"Metric type should be of type 'str': {self.type}"
            )
        elif not isinstance(self.value, (int, float, dict)):
            raise TypeError(
                f"Metric value must be of type 'int', 'float' or 'dict': {self.value}"
            )
        elif not isinstance(self.parameters, dict):
            raise TypeError(
                f"Metric parameters must be of type 'dict[str, Any]': {self.parameters}"
            )
        elif not all([isinstance(k, str) for k in self.parameters.keys()]):
            raise TypeError(
                f"Metric parameter dictionary should only have keys with type 'str': {self.parameters}"
            )

    @classmethod
    def precision(
        cls,
        value: float,
        label: str,
    ):
        """
        Precision metric for a specific class label.

        Precision is calulated using the number of true-positive pixels divided by
        the sum of all true-positive and false-positive pixels.

        Parameters
        ----------
        value : float
            The computed precision value.
        label : str
            The label for which the precision is calculated.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Precision.value,
            value=value,
            parameters={
                "label": label,
            },
        )

    @classmethod
    def recall(
        cls,
        value: float,
        label: str,
    ):
        """
        Recall metric for a specific class label.

        Recall is calulated using the number of true-positive pixels divided by
        the sum of all true-positive and false-negative pixels.

        Parameters
        ----------
        value : float
            The computed recall value.
        label : str
            The label for which the recall is calculated.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.Recall.value,
            value=value,
            parameters={
                "label": label,
            },
        )

    @classmethod
    def f1_score(
        cls,
        value: float,
        label: str,
    ):
        """
        F1 score for a specific class label.

        Parameters
        ----------
        value : float
            The computed F1 score.
        label : str
            The label for which the F1 score is calculated.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.F1.value,
            value=value,
            parameters={
                "label": label,
            },
        )

    @classmethod
    def iou(
        cls,
        value: float,
        label: str,
    ):
        """
        Intersection over Union (IOU) ratio for a specific class label.

        Parameters
        ----------
        value : float
            The computed IOU ratio.
        label : str
            The label for which the IOU is calculated.

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.IOU.value,
            value=value,
            parameters={
                "label": label,
            },
        )

    @classmethod
    def mean_iou(cls, value: float):
        """
        Mean Intersection over Union (mIOU) ratio.

        The mIOU value is computed by averaging IOU over all labels.

        Parameters
        ----------
        value : float
            The mIOU value.

        Returns
        -------
        Metric
        """
        return cls(type=MetricType.mIOU.value, value=value, parameters={})

    @classmethod
    def accuracy(cls, value: float):
        """
        Accuracy metric computed over all labels.

        Parameters
        ----------
        value : float
            The accuracy value.

        Returns
        -------
        Metric
        """
        return cls(type=MetricType.Accuracy.value, value=value, parameters={})

    @classmethod
    def confusion_matrix(
        cls,
        confusion_matrix: dict[
            str,  # ground truth label value
            dict[
                str,  # prediction label value
                dict[str, float],  # iou
            ],
        ],
        hallucinations: dict[
            str,  # prediction label value
            dict[str, float],  # pixel ratio
        ],
        missing_predictions: dict[
            str,  # ground truth label value
            dict[str, float],  # pixel ratio
        ],
    ):
        """
        The confusion matrix and related metrics for semantic segmentation tasks.

        This class encapsulates detailed information about the model's performance, including correct
        predictions, misclassifications, hallucinations (false positives), and missing predictions
        (false negatives). It provides counts for each category to facilitate in-depth analysis.

        Confusion Matrix Format:
        {
            <ground truth label>: {
                <prediction label>: {
                    'iou': <float>,
                },
            },
        }

        Hallucinations Format:
        {
            <prediction label>: {
                'iou': <float>,
            },
        }

        Missing Predictions Format:
        {
            <ground truth label>: {
                'iou': <float>,
            },
        }

        Parameters
        ----------
        confusion_matrix : dict
            Nested dictionaries representing the Intersection over Union (IOU) scores for each
            ground truth label and prediction label pair.
        hallucinations : dict
            Dictionary representing the pixel ratios for predicted labels that do not correspond
            to any ground truth labels (false positives).
        missing_predictions : dict
            Dictionary representing the pixel ratios for ground truth labels that were not predicted
            (false negatives).

        Returns
        -------
        Metric
        """
        return cls(
            type=MetricType.ConfusionMatrix.value,
            value={
                "confusion_matrix": confusion_matrix,
                "hallucinations": hallucinations,
                "missing_predictions": missing_predictions,
            },
            parameters={},
        )
